#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Activity Report views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework_extensions.mixins import DetailSerializerMixin
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import VLECourseActivityReportExtendedSerializer
from tesla_ce.apps.api.v2.serializers import VLECourseActivityReportSerializer
from tesla_ce.models import ReportActivity


# pylint: disable=too-many-ancestors
class VLECourseActivityReportViewSet(NestedViewSetMixin, DetailSerializerMixin, viewsets.ReadOnlyModelViewSet):
    """
    API endpoint that allows view activity reports.
    """
    model = ReportActivity
    queryset = ReportActivity.objects
    serializer_class = VLECourseActivityReportSerializer
    permission_classes = [
        permissions.VLEPermission
    ]
    serializer_detail_class = VLECourseActivityReportExtendedSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    search_fields = ['learner__first_name', 'learner__last_name', 'learner__email',
                     'identity_level', 'authorship_level', 'cheating_level']



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Base Command for TeSLA CE """
import os

from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

from tesla_ce.client import Client
from tesla_ce.client import ConfigManager


class TeslaCommand(BaseCommand):
    """ Base Command class for TeSLA CE """

    # TeSLA CE client instance
    _client: Client = None

    # Configuration file
    _conf_file = None

    # Execution options
    _options = None

    # Enable if configuration file is required
    _conf_file_required = False

    @staticmethod
    def get_client(config_file=None, options=None):
        """
            Create an instance of the client
            :param config_file: Path to the configuration file
            :param options: Provided options to the command
            :return: Instance of client
        """
        return Client(config=settings.TESLA_CONFIG, use_vault=False, use_env=False, enable_management=True)

    @staticmethod
    def get_config_file(options=None):
        """
            Get the configuration file
            :param options: Provided options to the command
            :return: Instance of client
        """
        # Get value in environment
        config_file = settings.TESLA_CONFIG.config.get('TESLA_CONFIG_FILE')
        if config_file is None:
            # Check default paths
            config_file = ConfigManager.find_config_file()

        return config_file

    @property
    def client(self):
        """
            Access to the TeSLA CE client
            :return: Instance of the client
        """
        if self._client is None:
            self._client = self.get_client(self.conf_file, self._options)
        return self._client

    @property
    def conf_file(self):
        """
            Access to the configuration file
            :return: Path to the configuration file
        """
        if self._conf_file is None:
            self._conf_file = self.get_config_file(self._options)
        return self._conf_file

    def print_version(self):
        """
            Print current software version
        """
        self.stdout.write('TeSLA CE version {}'.format(self.client.version))

    def check_configuration_file(self):
        """
            Check if the defined configuration file exists
        """
        # Check if file exists or not
        if self._conf_file is None:
            if self._conf_file_required:
                raise CommandError('Configuration file not found')
            else:
                self.stdout.write('Configuration file not found: {}'.format(self.style.WARNING('[WARINING]')))
        else:
            if os.path.exists(self.conf_file):
                self.stdout.write('Reading configuration from {}: {}'.format(self.conf_file,
                                                                             self.style.SUCCESS('[OK]')))
            else:
                if self._conf_file_required:
                    self.stdout.write('Reading configuration from {}: {}'.format(self.conf_file,
                                                                                 self.style.ERROR('[ERROR]')))
                    raise CommandError('Configuration file not found')
                else:
                    self.stdout.write('Error reading configuration from {}: {}'.format(self.conf_file,
                                                                                       self.style.WARINING('[WARNING]'))
                                      )

    def handle(self, *args, **options):
        """
            Perform command actions
            :param args: List of arguments
            :param options: List of named options
        """
        # Store options
        self._options = options

        # Print current version
        self.print_version()

        # Check configuration file
        self.check_configuration_file()

        # Perform custom actions
        self.custom_handle()

    def custom_handle(self):
        """
            Custom actions for each command
        """
        # This method is implemented in the commands
        pass


class TeslaConfigCommand(TeslaCommand):
    """ Base Configuration Command class for TeSLA CE """

    _conf_file_required = True

    @staticmethod
    def get_config_file(options=None):
        """
            Get the configuration file
            :param options: Provided options to the command
            :return: Instance of client
        """
        # Search for configuration file
        if options is not None and options['local']:
            config_file = 'tesla-ce.cfg'
        else:
            config_file = ConfigManager.find_config_file()

        return config_file

    @staticmethod
    def get_client(config_file=None, options=None):
        """
            Create an instance of the client
            :param config_file: Path to the configuration file
            :param options: Provided options to the command
            :return: Instance of client
        """
        return Client(config_file=config_file, use_vault=False, use_env=False, enable_management=True)

    def _get_configuration_report(self, allow_invalid=False):
        """
            Check if current configuration is valid
            :return: Configuration report
        """
        # Check the configuration
        self.stdout.write('Checking current configuration')
        report = self.client.check_configuration()
        if report['valid']:
            self.stdout.write(
                self.style.SUCCESS('Configuration file is {}.'.format(self.style.SUCCESS('valid')))
            )
        else:
            self.stdout.write(
                self.style.SUCCESS('Configuration file is {}.'.format(self.style.ERROR('not valid')))
            )
            if not allow_invalid:
                self.stdout.write(self.style.ERROR('Invalid configuration. Use --check option for details'))
                raise CommandError('Invalid configuration')

            for service in report:
                if isinstance(report[service], dict) and not report[service]['valid']:
                    self.stdout.write('Checking service {}: {}'.format(service, self.style.ERROR('[ERROR]')))
                    for err in report[service]['errors']:
                        self.stdout.write(self.style.ERROR('\t- {}'.format(err)))
            self.stdout.write(
                self.style.ERROR('Configuration errors. Check provided errors.'))

        return report


class TeslaDeployCommand(TeslaCommand):
    """ Base Deployment Command class for TeSLA CE """

    def add_arguments(self, parser):
        """
            Define custom arguments for this command

            :param parser: Input command parser instance
        """
        parser.add_argument(
            '--mode',
            choices=['swarm'],
            help='Type of deployment',
        )
        parser.add_argument(
            '--out',
            default='./deploy',
            help='Directory where configuration files will be written.',
        )



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" API Views for TeSLA eAssessment Portal (TEP) """
from rest_framework.views import APIView
from rest_framework.views import Response
from rest_framework.views import status

from tesla_ce import get_default_client
from tesla_ce.apps.api.auth.serializers import AuthTokenPairSerializer
from tesla_ce.models import Institution
from tesla_ce.models import Learner
from ..serializers.tep import LearnerEnrolment


class LearnerEnrolmentView(APIView):
    """
        Get enrolment data
    """
    def get(self, request, learner_id, format=None):
        # Use default institution
        try:
            institution = Institution.objects.get(pk=1)
        except Institution.DoesNotExist:
            return Response(status=status.HTTP_404_NOT_FOUND)
        # Search for the learner
        try:
            learner = Learner.objects.get(institution=institution, learner_id=learner_id)
        except Learner.DoesNotExist:
            return Response(status=status.HTTP_404_NOT_FOUND)
        return Response(LearnerEnrolment(learner).data)


class LearnerEnrolmentTokenView(APIView):
    """
        Get enrolment token
    """
    def get(self, request, learner_id, format=None):
        # Use default institution
        try:
            institution = Institution.objects.get(pk=1)
        except Institution.DoesNotExist:
            return Response(status=status.HTTP_404_NOT_FOUND)
        # Search for the learner
        try:
            learner = Learner.objects.get(institution=institution, learner_id=learner_id)
        except Learner.DoesNotExist:
            return Response(status=status.HTTP_404_NOT_FOUND)

        tokens = get_default_client().get_learner_token_pair(learner, ['/lapi/v1/enrolment/*'])

        return Response(AuthTokenPairSerializer(tokens).data)



# Generated by Django 3.0.6 on 2020-05-20 18:09

import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0004_launcher'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='assessmentsession',
            name='connector',
        ),
        migrations.CreateModel(
            name='AssessmentSessionData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('connector', models.FileField(help_text='Connector JS file for this session', upload_to='')),
                ('data', models.FileField(help_text='Data for this session', upload_to='')),
                ('session', models.ForeignKey(help_text='Related assessment session', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.AssessmentSession')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Learner api serialize module."""
from rest_framework import serializers
from rest_framework import validators

from tesla_ce.apps.api.utils import BodySerializer
from tesla_ce.models import InformedConsent
from tesla_ce.models import Institution
from tesla_ce.models import InstitutionUser
from tesla_ce.models import Learner
from tesla_ce.models import User
from .institution import InstitutionSerializer


class InstitutionLearnerConsentSerializer(serializers.ModelSerializer):
    """Learner Informed Consent serialize model module."""

    status = serializers.CharField(read_only=True)

    class Meta:
        model = InformedConsent
        fields = ["version", "status"]


class InstitutionLearnerSerializer(serializers.ModelSerializer):
    """Learner serialize model module."""

    institution = InstitutionSerializer(read_only=True)
    username = serializers.CharField(read_only=True,
                                     validators=[
                                         validators.UniqueValidator(queryset=User.objects.all())
                                     ])
    consent = InstitutionLearnerConsentSerializer(read_only=True, allow_null=True, default=None)
    consent_accepted = serializers.DateTimeField(read_only=True)
    consent_rejected = serializers.DateTimeField(read_only=True)
    learner_id = serializers.UUIDField(read_only=True)
    last_login = serializers.DateTimeField(read_only=True)
    send = serializers.JSONField(read_only=True)
    email = serializers.EmailField(validators=[
        validators.UniqueValidator(queryset=User.objects.all())
    ])
    institution_id = serializers.HiddenField(default=None, allow_null=True)
    ic_status = serializers.CharField(read_only=True)
    login_allowed = serializers.BooleanField(read_only=True, allow_null=False, default=False)

    class Meta:
        model = Learner
        exclude = ["password", "is_superuser", "is_staff", "is_active", "date_joined", "groups", "user_permissions"]
        validators = [serializers.UniqueTogetherValidator(
            queryset=InstitutionUser.objects.all(),
            fields=['institution_id', 'uid']
        ), ]

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['institution_id'] = self.context['view'].kwargs['parent_lookup_institution_id']
        attrs['username'] = attrs['email']

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)

    def validate_email(self, value):
        """
            Validate the given email
            :param value: Email for the learner
            :type value: str
            :return: Validated value
            :rtype: str
        """
        try:
            institution = Institution.objects.get(id=self.context['view'].kwargs['parent_lookup_institution_id'])
        except Institution.DoesNotExist:
            raise serializers.ValidationError('Invalid institution')
        if institution.mail_domain is not None and len(institution.mail_domain) > 0 \
                and not value.endswith(institution.mail_domain):
            raise serializers.ValidationError('Provided email does not belong to institution domain')
        return value

    def create(self, validated_data):
        """
            Create a new instance from provided data

            :param validated_data: Validated data from request
            :return: Created instance
            :rtype: Learner
        """
        instance = super().create(validated_data)
        instance.set_unusable_password()
        instance.save()
        return instance


class InstitutionLearnerICBodySerializer(BodySerializer):
    """ Body parser for Informed Consent acceptance """
    version = serializers.CharField(required=True)


class InstitutionLearnerDetailSerializer(InstitutionLearnerSerializer):
    enrolment_status = serializers.JSONField(read_only=True)




#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" ActivityInstrument model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from tesla_ce.apps.api.utils import JSONField
from .activity import Activity
from .base_model import BaseModel
from .instrument import Instrument

SENSOR = (
    ('camera', _('Capture images from webcam')),
    ('keyboard', _('Capture keyboard events')),
    ('microphone', _('Capture audio')),
    ('activity', _('Submission of the delivered activity')),
)


class ActivityInstrument(BaseModel):
    """ ActivityInstrument model. """
    activity = models.ForeignKey(Activity, null=False, blank=False,
                                 on_delete=models.CASCADE,
                                 help_text=_('Activity.'),
                                 related_name='configuration')

    instrument = models.ForeignKey(Instrument, null=False, blank=False,
                                   on_delete=models.CASCADE,
                                   help_text=_('Activity.'))

    alternative_to = models.ForeignKey("ActivityInstrument", null=True, blank=False,
                                       on_delete=models.CASCADE,
                                       help_text=_('Primary instrument to be used.'))

    active = models.BooleanField(null=False, blank=False,
                                 help_text="Is instrument active?")

    options = models.TextField(null=True, blank=True,
                               help_text="Instrument options?")

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    @property
    def is_primary(self):
        """
            Whether this instrument is the primary option or is an alternative option
            :return: True if it is primary instrument
            :rtype: bool
        """
        return self.alternative_to is None

    @property
    def is_enabled(self):
        """
            Whether this instrument is enabled or not
            :return: True if it is enabled and false if it is disabled
            :rtype: bool
        """
        if self.active:
            return True
        if self.is_primary:
            return self.active

        return self.alternative_to.is_enabled

    def __repr__(self):
        return "<ActivityInstrument(activity_id='%r', instrument_id='%r', " \
               "is_primary='%r', active='%r')>" % (
                   self.activity.id, self.instrument,
                   self.is_primary, self.active)

    def get_options(self):
        """
            Whether this instrument is the primary option or is an alternative option
            :return: True if it is primary instrument
            :rtype: bool
        """
        if self.options is not None:
            json_field = JSONField()
            return json_field.to_representation(self.options)

        return None

    def get_sensors(self):
        """
            Return the list of required sensors for this instrument
            :return: List of sensors
        """
        # If this instrument is disabled, no sensor is required
        if not self.active:
            return []

        # Get configuration options for this activity
        options = self.get_options()

        # Get sensors according to options
        sensors = []
        if self.instrument.acronym == 'fr':
            if 'online' in options and options['online']:
                sensors.append('camera')
            if 'offline' in options and options['offline']:
                sensors.append('activity')
        elif self.instrument.acronym == 'ks':
            sensors.append('keyboard')
        elif self.instrument.acronym == 'vr':
            if 'online' in options and options['online']:
                sensors.append('microphone')
            if 'offline' in options and options['offline']:
                sensors.append('activity')
        elif self.instrument.acronym == 'fa':
            sensors.append('activity')
        elif self.instrument.acronym == 'plag':
            sensors.append('activity')

        return sensors



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
# TeSLA CE Authentication Middleware
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from rest_framework.authentication import BaseAuthentication
from rest_framework.authentication import get_authorization_header
from rest_framework.exceptions import AuthenticationFailed

from tesla_ce import models
from tesla_ce.client import TeslaAuthException
from tesla_ce.client import get_default_client
from tesla_ce.models.user import AuthenticatedModule
from tesla_ce.models.user import AuthenticatedUser
from tesla_ce.models.user import UnauthenticatedUser


def _get_token(request):
    """
        Extract the authentication token from a request

        :param request: The request
        :type request: django.http.request
        :return: The token or None if it is not provided
        :rtype: str
    """
    # Get the Authentication header
    auth_header = request.headers.get('authorization')

    # Check if it contains a TeSLA JWT token
    if auth_header.startswith(settings.TESLA_JWT_TOKEN):
        token = auth_header.split()
        if len(token) == 2:
            token = token[1]
            return token
    return None


def _is_allowed_url(request, scopes, filters=None):
    """
        Check if current request is allowed

        :param request: The request to be checked
        :type request: django.http.request
        :param scopes: List o allowed scopes
        :type scopes: list
        :param filters: List of filtered scopes
        :type filters: list
        :return: True if it is allowed or False otherwise
        :rtype: bool
    """
    return True


def is_authenticated_request(request):
    """
        Check if a request is authenticated or not

        :param request: A request
        :type request: django.
        :return: None if it is not authenticated or the payload if it is valid
        :rtype: dict
    """
    # Get the token
    token = _get_token(request)

    if token is not None:
        payload = get_default_client().validate_token(token)
        if _is_allowed_url(request, payload.get('scope'), payload.get('filters')):
            return payload

    return None


def _authenticate(request, token):
    # Validate the token
    user = UnauthenticatedUser()
    try:
        validation = get_default_client().validate_token(token=token)
        if validation['valid']:
            user = UnauthenticatedUser()
    except TeslaAuthException:
        pass

    # Set the user to the request
    request.user = user


class JWTAuthentication(BaseAuthentication):
    """
        Authentication class for DJango Rest Framework
    """
    # Client used to verify tokens
    _client = get_default_client()

    def authenticate(self, request):
        """
            Authenticate the request and return a two-tuple of (user, token).

            :return: A two-tuple of (user, token).
            :rtype: tuple
        """
        # Get the Authentication header
        auth = get_authorization_header(request).split()

        if len(auth) == 1:
            msg = _('Invalid token header. No credentials provided.')
            raise AuthenticationFailed(msg)

        if len(auth) > 2:
            msg = _('Invalid token header. Token string should not contain spaces.')
            raise AuthenticationFailed(msg)

        if len(auth) == 0:
            return UnauthenticatedUser(), None

        try:
            token = auth[1].decode()
        except UnicodeError as eue:
            msg = _('Invalid token header. Token string should not contain invalid characters.')
            raise AuthenticationFailed(msg) from eue

        return self.authenticate_credentials(token)

    def _get_module_model(self, payload):
        """
            Get the corresponding module model from payload information
            :param payload: Extracted payload
            :return: Module instance
        """
        user = AuthenticatedModule()
        if payload['type'] == 'provider':
            try:
                user.module = models.Provider.objects.get(pk=payload['pk'])
            except models.Provider.DoesNotExist as dne:
                raise AuthenticationFailed(_('Invalid provider.')) from dne
        elif payload['type'] == 'vle':
            try:
                user.module = models.VLE.objects.get(pk=payload['pk'])
            except models.VLE.DoesNotExist as dne:
                raise AuthenticationFailed(_('Invalid VLE.')) from dne
        else:
            raise AuthenticationFailed(_('Invalid module type.'))

        user.id = None
        user.type = payload['type']
        user.pk = payload['pk']
        user.username = payload['sub']

        return user

    def _get_user_model(self, payload):
        """
            Get the corresponding user model from payload information
            :param payload: Extracted payload
            :return: User instance
        """
        if payload['group'] == 'learners':
            try:
                user = models.Learner.objects.get(learner_id=payload['sub'])
            except models.Learner.DoesNotExist as dne:
                raise AuthenticationFailed(_('Invalid learner.')) from dne
        elif payload['group'] == 'users':
            if payload['type'] == 'admin':
                try:
                    user = models.User.objects.get(pk=payload['pk'])
                except models.User.DoesNotExist as dne:
                    raise AuthenticationFailed(_('Invalid user.')) from dne
            else:
                try:
                    user = models.InstitutionUser.objects.get(uid=payload['sub'])
                except models.InstitutionUser.DoesNotExist:
                    try:
                        user = models.InstitutionUser.objects.get(id=payload['pk'])
                    except models.InstitutionUser.DoesNotExist as dne:
                        raise AuthenticationFailed(_('Invalid user.')) from dne
        elif payload['group'].startswith('module_'):
            user = self._get_module_model(payload)
        else:
            raise AuthenticationFailed(_('Invalid payload group.'))
        return user

    def authenticate_credentials(self, token):
        """
            Check provided token and perform authentication

            :param token: The JWT token provided
            :type token: str
            :return: A two-tuple of (user, token).
            :rtype: tuple
        """
        user = UnauthenticatedUser()
        try:
            payload = self._client.validate_token(token=token)
            if len(settings.DATABASES) > 0 and settings.DATABASES['default']['ENGINE'] != 'django.db.backends.dummy':
                user = self._get_user_model(payload)
            else:
                user = AuthenticatedUser()
                user.type = payload['type']
                user.allowed_scopes = payload['scope']
                user.sub = payload['sub']
                user.pk = payload['pk']
                user.filters = payload['filters']

        except TeslaAuthException as exc:
            raise AuthenticationFailed(_('Invalid token: {}'.format(exc))) from exc

        if not user.is_active:
            raise AuthenticationFailed(_('User inactive or deleted.'))

        return user, token

    def authenticate_header(self, request):
        """
            Return a string to be used as the value of the `WWW-Authenticate`
            header in a `401 Unauthenticated` response, or `None` if the
            authentication scheme should return `403 Permission Denied` responses.
        """
        return "JWT"


def tesla_token_auth_middleware(get_response):
    """
        TeSLA Authentication Middleware.
        :param get_response:
        :return:
    """

    def middleware(request):

        # Extract the token
        token = _get_token(request)

        # Authenticate the request
        if token is not None:
            # Get user from token
            _authenticate(request, token)
            # Check if user is allowed for this request
            #if request.user.is_allowed(request):
            #    response = get_response(request)
            # response = get_response(request)

        response = get_response(request)
        # Code to be executed for each request/response after
        # the view is called.

        return response

    return middleware


class DebugJWTAuthentication(JWTAuthentication):
    """
        Authentication class used for debug purposes
    """

    def authenticate(self, request):
        """
            Authenticate the request and return a two-tuple of (user, token).

            :return: A two-tuple of (user, token).
            :rtype: tuple
        """
        # Get the Authentication header
        auth = get_authorization_header(request).split()

        if len(auth) == 0:
            return self.authenticate_credentials(None)

        return super().authenticate(request)

    def authenticate_credentials(self, token):
        """
            Check provided token and perform authentication

            :param token: The JWT token provided
            :type token: str
            :return: A two-tuple of (user, token).
            :rtype: tuple
        """
        if settings.DEBUG_AUTHENTICATION_OBJECT is None:
            return UnauthenticatedUser(), None

        obj_type = settings.DEBUG_AUTHENTICATION_OBJECT[0]

        payload = {
            "sub": settings.DEBUG_AUTHENTICATION_OBJECT[1],
            "pk": settings.DEBUG_AUTHENTICATION_OBJECT[1],
            "allowed_scopes": "",
            "filters": ""
        }
        if obj_type == "admin":
            payload['group'] = "users"
            payload['type'] = "admin"
        elif obj_type == "user":
            payload['group'] = "users"
            payload['type'] = "user"
        elif obj_type == "vle":
            payload['group'] = "module_vle_{}".format(payload['pk'])
            payload['type'] = "vle"
        elif obj_type == "provider":
            payload['group'] = "module_provider_{}".format(payload['pk'])
            payload['type'] = "provider"
        else:
            raise AuthenticationFailed(_('Invalid user credentials.'))

        if len(settings.DATABASES) > 0:
            user = self._get_user_model(payload)
        else:
            user = AuthenticatedUser()
            user.type = payload['type']
            user.allowed_scopes = payload['allowed_scopes']
            user.sub = payload['sub']
            user.pk = payload['pk']
            user.filters = payload['filters']

        return user, None



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Histogram of learner results for a provider model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .instrument import Instrument
from .learner import Learner


class HistogramLearnerInstrument(BaseModel):
    """ Histogram for instrument results for a Learner model. """

    learner = models.ForeignKey(Learner, null=False, blank=False, on_delete=models.CASCADE,
                                help_text=_('Related Learner.'))

    instrument = models.ForeignKey(Instrument, null=False, blank=False, on_delete=models.CASCADE,
                                   help_text=_('Related Instrument.'))

    b0 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 0 to 9%'))
    b1 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 10% to 19%'))
    b2 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 20% to 29%'))
    b3 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 30% to 39%'))
    b4 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 40% to 49%'))
    b5 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 50% to 59%'))
    b6 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 60% to 69%'))
    b7 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 70% to 79%'))
    b8 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 80% to 89%'))
    b9 = models.BigIntegerField(null=False, default=0, help_text=_('Bin containing values from 90% to 100%'))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('learner', 'instrument'),)

    def __repr__(self):
        return "<HistogramLearnerInstrument(id='%r', learner_id='%r', instrument_id='%r')>" % (
            self.id, self.learner_id, self.instrument_id)



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE Activity use case test """
import pytest

from tesla_ce.models.learner import get_missing_enrolment

from . import utils


@pytest.mark.django_db(transaction=False)
def test_activity_case_complete(rest_api_client, user_global_admin):
    """
        Complete case use for TeSLA
        :param rest_api_client: API client
        :param user_global_admin: Global administrator object
    """

    # A global administrator register the providers
    providers = utils.global_admin.api_register_providers(user_global_admin)

    # A global administrator creates a new institution
    institution = utils.global_admin.api_create_institution(user_global_admin)

    # A global administrator creates a new user for the institution and assign administration rights
    inst_admin = utils.global_admin.api_create_institution_admin(user_global_admin, institution)

    # Institution administrator creates a new legal admin
    legal_admin = utils.inst_admin.api_create_institution_legal_admin(inst_admin)

    # Institution administrator creates a new SEND admin
    send_admin = utils.inst_admin.api_create_institution_send_admin(inst_admin)

    # A legal administrator of the institution creates the Informed Consent using the API
    utils.inst_admin.api_create_ic(legal_admin, '1.0.0')

    # A SEND administrator of the institution defines the SEND categories using the API
    ks_id = providers['ks']['instrument']['id']
    send_category = utils.inst_admin.api_create_send_categories(send_admin, [ks_id])

    # Institution enables direct learners and instructors registration by VLE
    utils.inst_admin.api_enable_direct_registration_vle(inst_admin)

    # Institution administrator register a new VLE
    vle = utils.inst_admin.api_register_vle(inst_admin)

    # The VLE creates a new course
    course = utils.vle.vle_create_course(vle)

    # The VLE enrolls course learners and instructors
    instructors, learners = utils.vle.vle_enrol_users(vle, course)

    # The VLE creates a new activity
    activity = utils.vle.vle_create_activity(vle, course)

    # The VLE creates a launcher for the instructor
    instructor_launcher = utils.vle.vle_create_launcher(vle, instructors[0])

    # An instructor configures the activity using the API
    utils.instructor.api_configure_activity(instructor_launcher, activity, providers)

    for learner in learners:
        # VLE check the status of the Informed Consent of the learner
        utils.vle.vle_check_learner_ic(vle, course, learner, missing=True)

        # The VLE fails to create an assessment session because IC is missing
        utils.vle.vle_create_assessment_session(vle, learner, activity, ic=False)

        # The VLE creates a launcher for the learner to accept IC
        launcher_ic = utils.vle.vle_create_launcher(vle, learner)

        # The learner accepts the IC using the API
        utils.learner.api_learner_accept_ic(launcher_ic)

        # VLE check the status of the Informed Consent of the learner
        utils.vle.vle_check_learner_ic(vle, course, learner, missing=False)

        # The VLE fails to create an assessment session because of enrolment
        utils.vle.vle_create_assessment_session(vle, learner, activity, ic=True, enrolment=False)

    # The SEND admin assigns send category to the learner
    utils.inst_admin.api_set_learner_send(send_admin, send_category, learners[1])

    # Invalidate cache to avoid refresh time due to send status change
    get_missing_enrolment.invalidate(learners[1]['id'], activity['id'])

    # The VLE creates a launcher for the learners to check enrolments
    launcher_enrol1 = utils.vle.vle_create_launcher(vle, learners[0])
    launcher_enrol2 = utils.vle.vle_create_launcher(vle, learners[1])

    # A learner check their enrolment status via API
    l1_enrolment = utils.learner.api_learner_enrolment(launcher_enrol1)
    assert len(l1_enrolment) == 0
    l2_enrolment = utils.learner.api_learner_enrolment(launcher_enrol2)
    assert len(l2_enrolment) == 0

    # The VLE creates a launcher for the learners to check enrolments required for activity
    launcher_act_enrol1 = utils.vle.vle_create_launcher(vle, learners[0])
    launcher_act_enrol2 = utils.vle.vle_create_launcher(vle, learners[1])

    # A learner check missing enrolments for an activity via API
    l1_missing_enrolment = utils.learner.api_learner_missing_enrolment(launcher_act_enrol1, activity, missing=True)
    assert providers['fr']['instrument']['id'] not in l1_missing_enrolment['instruments']  # Alternative instrument
    assert providers['ks']['instrument']['id'] in l1_missing_enrolment['instruments']  # Primary instrument
    assert providers['plag']['instrument']['id'] not in l1_missing_enrolment['instruments']  # No enrolment

    l2_missing_enrolment = utils.learner.api_learner_missing_enrolment(launcher_act_enrol2, activity, missing=True)
    assert providers['fr']['instrument']['id'] in l2_missing_enrolment['instruments']  # Enabled as primary is disabled
    assert providers['ks']['instrument']['id'] not in l2_missing_enrolment['instruments']  # Disabled by SEND
    assert providers['plag']['instrument']['id'] not in l2_missing_enrolment['instruments']  # No enrolment

    # The VLE checks the enrolment status for the learner (missing is expected as is new learner)
    missing1 = utils.vle.vle_create_assessment_session(vle, learners[0], activity, enrolment=False)
    assert providers['fr']['instrument']['id'] not in missing1['enrolments']['instruments']  # Alternative instrument
    assert providers['ks']['instrument']['id'] in missing1['enrolments']['instruments']  # Primary instrument
    assert providers['plag']['instrument']['id'] not in missing1['enrolments']['instruments']  # No enrolment
    missing2 = utils.vle.vle_create_assessment_session(vle, learners[1], activity, enrolment=False)
    assert providers['fr']['instrument']['id'] in missing2['enrolments']['instruments']  # Enabled as disabled primary
    assert providers['ks']['instrument']['id'] not in missing2['enrolments']['instruments']  # Disabled by SEND
    assert providers['plag']['instrument']['id'] not in missing2['enrolments']['instruments']  # No enrolment

    # The VLE creates a launcher for the learner to perform the enrolment
    launcher_enrol1 = utils.vle.vle_create_launcher(vle, learners[0])
    launcher_enrol2 = utils.vle.vle_create_launcher(vle, learners[1])

    # The learner perform enrolment for missing instruments, sending data using LAPI
    provider_validation_tasks = []
    provider_validation_tasks += utils.learner.api_lapi_perform_enrolment(
        launcher_enrol1, list(missing1['enrolments']['instruments'].keys())
    )
    provider_validation_tasks += utils.learner.api_lapi_perform_enrolment(
        launcher_enrol2, list(missing2['enrolments']['instruments'].keys())
    )

    # The VLE creates a launcher for the learners to check enrolments
    launcher_stats1 = utils.vle.vle_create_launcher(vle, learners[0])
    launcher_stats2 = utils.vle.vle_create_launcher(vle, learners[1])

    # The learner check the status of sent enrolment samples
    utils.learner.lapi_check_sample_status(launcher_stats1, 'PENDING')
    utils.learner.lapi_check_sample_status(launcher_stats2, 'PENDING')

    # The VLE creates a launcher for the learners to check enrolments
    launcher_enrol_pre_val1 = utils.vle.vle_create_launcher(vle, learners[0])
    launcher_enrol_pre_val2 = utils.vle.vle_create_launcher(vle, learners[1])

    # A learner check their enrolment status via API
    l1_enrolment_pre_val = utils.learner.api_learner_enrolment(launcher_enrol_pre_val1)
    assert len(l1_enrolment_pre_val) == 1
    assert l1_enrolment_pre_val[0]['instrument_id'] == providers['ks']['instrument']['id']
    assert len(l1_enrolment_pre_val[0]['not_validated']) == 1
    assert l1_enrolment_pre_val[0]['not_validated'][0]['provider_id'] == providers['ks']['id']
    assert l1_enrolment_pre_val[0]['not_validated'][0]['instrument_id'] == providers['ks']['instrument']['id']
    assert l1_enrolment_pre_val[0]['not_validated'][0]['count'] == providers['ks']['instrument']['id'] * 4
    assert len(l1_enrolment_pre_val[0]['pending']) == 0

    l2_enrolment_pre_val = utils.learner.api_learner_enrolment(launcher_enrol_pre_val2)
    assert len(l2_enrolment_pre_val) == 1
    assert l2_enrolment_pre_val[0]['instrument_id'] == providers['fr']['instrument']['id']
    assert len(l2_enrolment_pre_val[0]['not_validated']) == 2
    if l2_enrolment_pre_val[0]['not_validated'][0]['provider_id'] == providers['fr']['id']:
        assert l2_enrolment_pre_val[0]['not_validated'][0]['instrument_id'] == providers['fr']['instrument']['id']
        assert l2_enrolment_pre_val[0]['not_validated'][0]['count'] == providers['fr']['instrument']['id'] * 4
        assert l2_enrolment_pre_val[0]['not_validated'][1]['provider_id'] == providers['fr_def']['id']
        assert l2_enrolment_pre_val[0]['not_validated'][1]['instrument_id'] == providers['fr_def']['instrument']['id']
        assert l2_enrolment_pre_val[0]['not_validated'][1]['count'] == providers['fr_def']['instrument']['id'] * 4
    else:
        assert l2_enrolment_pre_val[0]['not_validated'][0]['provider_id'] == providers['fr_def']['id']
        assert l2_enrolment_pre_val[0]['not_validated'][0]['instrument_id'] == providers['fr_def']['instrument']['id']
        assert l2_enrolment_pre_val[0]['not_validated'][0]['count'] == providers['fr_def']['instrument']['id'] * 4
        assert l2_enrolment_pre_val[0]['not_validated'][1]['provider_id'] == providers['fr']['id']
        assert l2_enrolment_pre_val[0]['not_validated'][1]['instrument_id'] == providers['fr']['instrument']['id']
        assert l2_enrolment_pre_val[0]['not_validated'][1]['count'] == providers['fr']['instrument']['id'] * 4
    assert len(l2_enrolment_pre_val[0]['pending']) == 0

    # Providers validate samples in their queues
    validation_summary_tasks = utils.provider.provider_validate_samples(providers, provider_validation_tasks)

    # As we have deferred providers, some of the validations are postponed via notifications.
    notification_tasks = utils.worker.worker_send_notifications()
    assert len(notification_tasks) > 0

    # Providers process their notifications to generate final validation results
    validation_summary_tasks += utils.provider.provider_process_notifications(providers,
                                                                              notification_tasks,
                                                                              'validation')

    # The VLE creates a launcher for the learners to check enrolments
    launcher_enrol_post_val1 = utils.vle.vle_create_launcher(vle, learners[0])
    launcher_enrol_post_val2 = utils.vle.vle_create_launcher(vle, learners[1])

    # A learner check their enrolment status via API
    l1_enrolment_post_val = utils.learner.api_learner_enrolment(launcher_enrol_post_val1)
    assert len(l1_enrolment_post_val) == 1
    assert l1_enrolment_post_val[0]['instrument_id'] == providers['ks']['instrument']['id']
    assert len(l1_enrolment_post_val[0]['not_validated']) == 0
    assert l1_enrolment_post_val[0]['not_validated_count'] == 0
    assert len(l1_enrolment_post_val[0]['pending']) == 1
    assert l1_enrolment_post_val[0]['pending'][0]['provider_id'] == providers['ks']['id']
    assert l1_enrolment_post_val[0]['pending'][0]['instrument_id'] == providers['ks']['instrument']['id']
    assert abs(1.0 - float(l1_enrolment_post_val[0]['pending'][0]['pending_contribution'])) < 0.0001

    l2_enrolment_post_val = utils.learner.api_learner_enrolment(launcher_enrol_post_val2)
    assert len(l2_enrolment_post_val) == 1
    assert l2_enrolment_post_val[0]['instrument_id'] == providers['fr']['instrument']['id']
    assert len(l2_enrolment_post_val[0]['not_validated']) == 0
    assert l2_enrolment_post_val[0]['not_validated_count'] == 0
    assert len(l2_enrolment_post_val[0]['pending']) == 2
    if l2_enrolment_post_val[0]['pending'][0]['provider_id'] == providers['fr']['id']:
        assert l2_enrolment_post_val[0]['pending'][0]['instrument_id'] == providers['fr']['instrument']['id']
        assert abs(1.0 - float(l2_enrolment_post_val[0]['pending'][0]['pending_contribution'])) < 0.0001
        assert l2_enrolment_post_val[0]['pending'][1]['provider_id'] == providers['fr_def']['id']
        assert l2_enrolment_post_val[0]['pending'][1]['instrument_id'] == providers['fr_def']['instrument']['id']
        assert abs(1.0 - float(l2_enrolment_post_val[0]['pending'][1]['pending_contribution'])) < 0.0001
    else:
        assert l2_enrolment_post_val[0]['pending'][0]['provider_id'] == providers['fr_def']['id']
        assert l2_enrolment_post_val[0]['pending'][0]['instrument_id'] == providers['fr_def']['instrument']['id']
        assert abs(1.0 - float(l2_enrolment_post_val[0]['pending'][0]['pending_contribution'])) < 0.0001
        assert l2_enrolment_post_val[0]['pending'][1]['provider_id'] == providers['fr']['id']
        assert l2_enrolment_post_val[0]['pending'][1]['instrument_id'] == providers['fr']['instrument']['id']
        assert abs(1.0 - float(l2_enrolment_post_val[0]['pending'][1]['pending_contribution'])) < 0.0001

    # Worker compute validation summary from individual validations
    enrolment_tasks = utils.worker.worker_validation_summary(validation_summary_tasks)

    # The VLE creates a launcher for the learners to check the status of their enrolment samples
    launcher_stats1 = utils.vle.vle_create_launcher(vle, learners[0])
    launcher_stats2 = utils.vle.vle_create_launcher(vle, learners[1])

    # The learner check the status of sent enrolment samples
    utils.learner.lapi_check_sample_status(launcher_stats1, ['VALID', 'ERROR'])
    utils.learner.lapi_check_sample_status(launcher_stats2, ['VALID', 'ERROR'])

    # Worker distribute enrolment tasks among providers
    provider_enrolment_tasks = utils.worker.worker_enrol_learner(enrolment_tasks)

    # Provider perform learners enrolment
    utils.provider.provider_enrol_learners(providers, provider_enrolment_tasks)

    # As we have deferred providers, some of the enrolments are postponed via notifications.
    notification_tasks = utils.worker.worker_send_notifications()
    assert len(notification_tasks) > 0

    # Providers process their notifications to generate final enrolment results
    provider_enrolment_tasks += utils.provider.provider_process_notifications(providers, notification_tasks, 'enrolment')

    # The VLE creates a launcher for the learners to check enrolments
    launcher_enrol_end1 = utils.vle.vle_create_launcher(vle, learners[0])
    launcher_enrol_end2 = utils.vle.vle_create_launcher(vle, learners[1])

    # A learner check their enrolment status via API
    l1_enrolment_end = utils.learner.api_learner_enrolment(launcher_enrol_end1)
    assert len(l1_enrolment_end) == 1
    assert l1_enrolment_end[0]['instrument_id'] == providers['ks']['instrument']['id']
    assert len(l1_enrolment_end[0]['not_validated']) == 0
    assert len(l1_enrolment_end[0]['pending']) == 0
    assert abs(1.0 - float(l1_enrolment_end[0]['percentage__min'])) < 0.0001
    assert abs(1.0 - float(l1_enrolment_end[0]['percentage__max'])) < 0.0001
    assert l1_enrolment_end[0]['can_analyse__min']
    assert l1_enrolment_end[0]['can_analyse__max']
    assert l1_enrolment_end[0]['not_validated_count'] == 0

    l2_enrolment_end = utils.learner.api_learner_enrolment(launcher_enrol_end2)
    assert len(l2_enrolment_end) == 1
    assert l2_enrolment_end[0]['instrument_id'] == providers['fr']['instrument']['id']
    assert len(l2_enrolment_end[0]['not_validated']) == 0
    assert len(l2_enrolment_end[0]['pending']) == 0
    assert abs(1.0 - float(l2_enrolment_end[0]['percentage__min'])) < 0.0001
    assert abs(1.0 - float(l2_enrolment_end[0]['percentage__max'])) < 0.0001
    assert l2_enrolment_end[0]['can_analyse__min']
    assert l2_enrolment_end[0]['can_analyse__max']
    assert l2_enrolment_end[0]['not_validated_count'] == 0

    # Invalidate cache to avoid refresh time
    get_missing_enrolment.invalidate(learners[0]['id'], activity['id'])
    get_missing_enrolment.invalidate(learners[1]['id'], activity['id'])

    # The VLE creates an assessment session for a learner for the activity
    vle_options = {
        'floating_menu_initial_pos': 'top-right'
    }
    assessment_session1 = utils.vle.vle_create_assessment_session(vle, learners[0], activity, vle_options)
    assessment_session2 = utils.vle.vle_create_assessment_session(vle, learners[1], activity)

    # The learner perform the activity, sending information from sensors using the LAPI
    provider_verification_tasks = []
    tasks, activity_doc1 = utils.learner.lapi_learner_perform_activity(assessment_session1)
    provider_verification_tasks += tasks
    tasks, activity_doc2 = utils.learner.lapi_learner_perform_activity(assessment_session2)
    provider_verification_tasks += tasks

    # VLE send the activity
    provider_verification_tasks += utils.vle.vle_send_activity(vle, assessment_session1, activity_doc1, True)
    provider_verification_tasks += utils.vle.vle_send_activity(vle, assessment_session2, activity_doc2)

    # The VLE creates a launcher for the learners to check the status of their requests
    launcher_stats1 = utils.vle.vle_create_launcher(vle, learners[0])
    launcher_stats2 = utils.vle.vle_create_launcher(vle, learners[1])

    # The learner check the status of sent requests
    utils.learner.lapi_check_requests_status(launcher_stats1, 'VALID')
    utils.learner.lapi_check_requests_status(launcher_stats2, 'VALID')

    # Provider perform verification process on data collected during the activity
    reporting_tasks = utils.provider.provider_verify_request(providers, provider_verification_tasks)

    # As we have deferred providers, some of the verifications are postponed via notifications.
    notification_tasks = utils.worker.worker_send_notifications()
    assert len(notification_tasks) > 0

    # Providers process their notifications to generate final verification results
    reporting_tasks += utils.provider.provider_process_notifications(providers, notification_tasks, 'verification')

    # Worker process verification results to create the reports
    utils.worker.worker_create_reports(reporting_tasks)

    # Instructor review the results for the activity
    launcher_report = utils.vle.vle_create_launcher(vle, instructors[0])
    reports = utils.instructor.api_instructor_report(launcher_report, activity)
    assert len(reports) == 2
    for report in reports:
        assert report['identity_level'] == 2  # OK
        assert report['content_level'] == 2  # OK
        assert report['integrity_level'] == 1  # NO INFORMATION
        if report['learner']['id'] == learners[0]['id']:
            inst_ref = 2
        else:
            inst_ref = 1
        assert len(report['detail']) == 2
        assert len(report['detailed_report']['requests']) == (10 + 1)  # 10 from ks or fr and 1 for plagiarism
        if report['detail'][0]['instrument_id'] == 5:
            assert report['detail'][0]['identity_level'] == 1
            assert report['detail'][0]['content_level'] == 2
            assert report['detail'][0]['integrity_level'] == 1
            assert report['detail'][1]['instrument_id'] == inst_ref
            assert report['detail'][1]['identity_level'] == 2
            assert report['detail'][1]['content_level'] == 1
            assert report['detail'][1]['integrity_level'] == 1
        else:
            assert report['detail'][0]['instrument_id'] == inst_ref
            assert report['detail'][0]['identity_level'] == 2
            assert report['detail'][0]['content_level'] == 1
            assert report['detail'][0]['integrity_level'] == 1
            assert report['detail'][1]['instrument_id'] == 5
            assert report['detail'][1]['identity_level'] == 1
            assert report['detail'][1]['content_level'] == 2
            assert report['detail'][1]['integrity_level'] == 1

    # VLE get the results for the activity for integration
    reports_vle = utils.vle.vle_activity_report(vle, activity)
    assert len(reports_vle) == 2

    # Instructor review the results for the activity using filters
    launcher_report2 = utils.vle.vle_create_launcher(vle, instructors[0])
    reports2 = utils.instructor.api_instructor_report(launcher_report2,
                                                      activity,
                                                      filters='identity_level=2&content_level=2')
    assert len(reports2) == 2
    launcher_report3 = utils.vle.vle_create_launcher(vle, instructors[0])
    reports3 = utils.instructor.api_instructor_report(launcher_report3,
                                                      activity,
                                                      filters='identity_level=2')
    assert len(reports3) == 2
    launcher_report4 = utils.vle.vle_create_launcher(vle, instructors[0])
    reports4 = utils.instructor.api_instructor_report(launcher_report4,
                                                      activity,
                                                      filters='identity_level=3')
    assert len(reports4) == 0

    # Legal admin updates Informed consent status with minor change
    utils.inst_admin.api_create_ic(legal_admin, '1.0.1')

    # Learners IC status is still valid
    utils.vle.vle_check_learner_ic(vle, course, learners[0], missing=False)
    utils.vle.vle_check_learner_ic(vle, course, learners[1], missing=False)

    # Legal admin updates Informed consent status with major change
    utils.inst_admin.api_create_ic(legal_admin, '1.1.0')

    # Learners IC status is not valid
    utils.vle.vle_check_learner_ic(vle, course, learners[0], missing=True)
    utils.vle.vle_check_learner_ic(vle, course, learners[1], missing=True)

    # Learner accepts new IC
    launcher_new_ic1 = utils.vle.vle_create_launcher(vle, learners[0])
    utils.learner.api_learner_accept_ic(launcher_new_ic1)
    launcher_new_ic2 = utils.vle.vle_create_launcher(vle, learners[1])
    utils.learner.api_learner_accept_ic(launcher_new_ic2)

    # Learner IC status is valid again
    utils.vle.vle_check_learner_ic(vle, course, learners[0], missing=False)
    utils.vle.vle_check_learner_ic(vle, course, learners[1], missing=False)

    # The learner 1 continues with the activity => Report update

    # The learner 2 continues with the activity and providers detect some issue => Report update + alerts

    # The learner 2 rejects the informed consent

    # Data admin check the list of rejected informed consents

    # Data admin remove all data of a learner


    # The VLE closes the sessions
    utils.vle.vle_close_assessment_session(vle, assessment_session1)
    utils.vle.vle_close_assessment_session(vle, assessment_session2)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Request related tasks """
from . import alerts
from . import enrolment
from . import verification

__all__ = ['verification', 'enrolment', 'alerts']



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango command to register a new VLE to TeSLA system """
import json
import os

from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

from tesla_ce.client import Client


class Command(BaseCommand):
    help = 'Register a new VLE to TeSLA CE system'
    requires_system_checks = '__all__'

    def add_arguments(self, parser):
        parser.add_argument(
            '--type',
            choices=['moodle'],
            default='moodle',
            help='Type of VLE',
        )
        parser.add_argument(
            '--force-update',
            action='store_true',
            help='If the VLE already exist, update the information and generate new credentials',
        )
        parser.add_argument(
            '--url',
            help='URL of the VLE',
            default=None,
        )
        parser.add_argument(
            '--client-id',
            help='Client ID for LTI 1.3',
            default=None,
        )
        parser.add_argument(
            'name',
            help='Unique name for the VLE',
        )
        parser.add_argument(
            '--institution',
            help='The acronym of the institution this VLE belongs to',
            default=None,
        )

    def handle(self, *args, **options):

        # Check the configuration
        config_file = settings.TESLA_CONFIG.config.get('TESLA_CONFIG_FILE')
        if os.path.exists(config_file):
            self.stdout.write('Reading configuration from {}: {}'.format(config_file,
                                                                         self.style.SUCCESS('[OK]')))
        else:
            self.stdout.write('Reading configuration from {}: {}'.format(config_file,
                                                                         self.style.ERROR('[ERROR]')))
            raise CommandError('Configuration file not found')

        # Create a client
        client = Client(config=settings.TESLA_CONFIG, use_vault=False, use_env=False, enable_management=True)

        # Get arguments
        type = options['type']
        name = options['name']
        url = options['url']
        client_id = options['client_id']
        institution_acronym = options['institution']
        force_update = options['force_update']

        # Register the new VLE
        vle_info = client.register_vle(type, name, url, institution_acronym, client_id, force_update)

        self.stdout.write('VLE registered. Use this configuration: {}'.format(
            json.dumps(vle_info, indent=4, sort_keys=True),
            self.style.SUCCESS('[OK]'))
        )



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Exception definition module"""
# Initialize Sentry Error tracker
import os

import sentry_sdk
from sentry_sdk import capture_exception
from sentry_sdk.integrations.django import DjangoIntegration

# Initialize the Sentry Error Tracker
if os.getenv('SENTRY_ENABLED') in ['1', 1, 'True', 'yes', 'true'] and os.getenv('SENTRY_DSN') is not None:
    sentry_sdk.init(
        os.getenv('SENTRY_DSN'),
        max_breadcrumbs=50,
        debug=os.getenv('DEBUG', '0') in ['1', 1, 'True', 'yes', 'true'],
        release=open(os.path.join(os.path.dirname(__file__), 'data', 'VERSION'), 'r').read(),
        environment=os.getenv('SENTRY_ENVIRONMENT', 'production'),
        server_name=os.getenv('SENTRY_SERVER_NAME', None),
        integrations=[DjangoIntegration()],
        attach_stacktrace=True,

        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for performance monitoring.
        # We recommend adjusting this value in production,
        traces_sample_rate=1.0,

        # If you wish to associate users to errors (assuming you are using
        # django.contrib.auth) you may enable sending PII data.
        send_default_pii=True
    )


def tesla_report_exception(exception=None):
    """
        Send given exception to the Sentry Error Tracking system
        :param exception: Exception (optional)
        :return: Issue id or None if tracking is not enabled
    """
    if os.getenv('SENTRY_ENABLED') in ['1', 1, 'True', 'yes', 'true'] and os.getenv('SENTRY_DSN') is not None:
        return capture_exception(exception)
    return None


class TeslaException(Exception):
    """ Base class to raise exceptions """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        tesla_report_exception(self)


class TeslaConfigException(TeslaException):
    """ Class to raise configuration exceptions """
    pass


class TeslaVaultException(TeslaException):
    """ Class to raise vault related exceptions """
    pass


class TeslaAuthException(TeslaException):
    """ Class to raise authentication related exceptions """
    pass


class TeslaDatabaseException(TeslaException):
    """ Class to raise database related exceptions """
    pass


class TeslaDeploymentException(TeslaException):
    """ Class to raise deployment related exceptions """
    pass


class TeslaStorageException(TeslaException):
    """ Class to raise storage related exceptions """
    pass


class TeslaMissingICException(TeslaException):
    """ Class to raise exception when a learner does not have a valid informed consent """
    pass


class TeslaInvalidICException(TeslaException):
    """ Class to raise exception when a learner have an invalid informed consent """
    pass


class TeslaMissingEnrolmentException(TeslaException):
    """ Class to raise exception when required enrolments are missing """
    pass



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Provider enrolment serializers module """
from django.core.files.base import ContentFile
from django.db import transaction
from django.utils import timezone
from rest_framework import serializers

from tesla_ce.models import Enrolment
from tesla_ce.models import Learner
from tesla_ce.models.enrolment import get_upload_path


class ProviderEnrolmentSerializer(serializers.ModelSerializer):
    """Enrolment sample serialize class."""

    learner_id = serializers.UUIDField(source='learner.learner_id')
    model = serializers.FileField(read_only=True)
    task_id = serializers.UUIDField(write_only=True, allow_null=False)
    is_locked = serializers.BooleanField(read_only=True)
    model_upload_url = serializers.SerializerMethodField(read_only=True)
    percentage = serializers.FloatField(required=False, allow_null=None, default=0)
    can_analyse = serializers.BooleanField(required=False, allow_null=None, default=False)
    model_total_samples = serializers.IntegerField(read_only=True)
    used_samples = serializers.ListField(write_only=True, allow_null=True, allow_empty=True, required=False, default=[],
                                         child=serializers.IntegerField())

    class Meta:
        model = Enrolment
        fields = ['id', 'learner_id', 'model', 'is_locked', 'task_id', 'model_upload_url', 'percentage',
                  'can_analyse', 'model_total_samples', 'used_samples']

    def get_model_upload_url(self, instance):
        if not self.context['request'].method == 'POST':
            return None
        pre_post = instance.model.storage.bucket.meta.client.generate_presigned_post(
            Bucket=instance.model.storage.bucket.name,
            Key=get_upload_path(instance, None)
        )

        """
        TODO: Move to presigned URL  => Check: https://boto3.amazonaws.com/v1/documentation/api/1.9.42/guide/s3.html#generating-presigned-urls
        params = {
            'Bucket': instance.model.storage.bucket.name,
            'Key': get_upload_path(instance, None)
        }
        url = instance.model.storage.bucket.meta.client.generate_presigned_url(
            'put_object', Params=params
        )
        """

        return pre_post

    def validate_learner_id(self, value):
        """
            Validate the provided learner UUID ID
            :param value: Learner UUID ID
            :type value: str
            :return: Validated value
            :rtype: str
        """
        try:
            learner = Learner.objects.get(learner_id=value)
        except Learner.DoesNotExist:
            raise serializers.ValidationError('Invalid learner')
        return learner.id

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['provider_id'] = self.context['view'].kwargs['parent_lookup_provider_id']
        attrs['learner_id'] = attrs['learner']['learner_id']

        # Check if model exists
        with transaction.atomic():
            # Get current model or create a new one
            try:
                model = Enrolment.objects.get(provider_id=attrs['provider_id'], learner_id=attrs['learner_id'])
            except Enrolment.DoesNotExist:
                model = Enrolment.objects.create(
                    provider_id=attrs['provider_id'],
                    learner_id=attrs['learner_id'],
                    percentage=0,
                    can_analyse=False
                )
                model.model.save(get_upload_path(model, None),  ContentFile('{}'.encode('utf-8')))
            # Check if model is locked
            if model.is_locked and str(model.locked_by) != str(attrs['task_id']):
                # If model lock have more than 5 hours, release previous lock
                time_threshold = timezone.now() - timezone.timedelta(hours=5)
                if model.locked_at >= time_threshold:
                    raise serializers.ValidationError('Model is locked')

            # Lock the model
            model.locked_by = attrs['task_id']
            model.locked_at = timezone.now()
            model.save()

            # Set model as current instance
            self.instance = model

        return super().validate(attrs)

    def update(self, instance, validated_data):
        if self.context['request'].method == 'PUT':
            # Update model metadata and unlock it
            if 'percentage' in validated_data:
                instance.percentage = validated_data['percentage']
            if 'can_analyse' in validated_data:
                instance.can_analyse = validated_data['can_analyse']
            if 'used_samples' in validated_data:
                # Update existing samples in validation
                instance.model_samples.set(validated_data['used_samples'])

            instance.locked_at = None
            instance.locked_by = None
            instance.save()
        return instance



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Course group model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .course import Course
from .institution import Institution


class CourseGroup(BaseModel):
    """ Course model. """
    institution = models.ForeignKey(Institution, null=False, blank=False,
                                    on_delete=models.CASCADE,
                                    help_text=_('Institution for this group.'))

    name = models.CharField(max_length=250, null=False, blank=False,
                            help_text=_("Group name."))

    description = models.TextField(null=True, blank=True,
                                   help_text=_("Group description."))

    parent = models.ForeignKey("self", null=True, blank=True,
                               on_delete=models.CASCADE,
                               related_name='children',
                               help_text=_('Parent group.'))

    courses = models.ManyToManyField(Course, blank=True,
                                     help_text=_('Courses in this group.'))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<CourseGroup(id='%r', institution_id='%r', name='%s')>" % (
            self.id, self.institution_id, self.name)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Instrument views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter

from tesla_ce.apps.api.v2.serializers import InstrumentSerializer
from tesla_ce.models import Instrument


# pylint: disable=too-many-ancestors
class InstitutionInstrumentViewSet(viewsets.ReadOnlyModelViewSet):
    """
    API endpoint that allows read instrument data
    """
    queryset = Instrument.objects.filter(enabled=True).order_by('acronym')
    serializer_class = InstrumentSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    # TODO: Filter instruments allowed by institution (new feature)
    '''
    filterset_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    search_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    '''



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" RequestResult views module """
from django.shortcuts import get_object_or_404
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework.mixins import UpdateModelMixin
from rest_framework.views import Response
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import ProviderVerificationRequestResultSerializer
from tesla_ce.models import RequestProviderResult


# pylint: disable=too-many-ancestors
class ProviderVerificationRequestResultViewSet(NestedViewSetMixin, UpdateModelMixin, viewsets.ReadOnlyModelViewSet):
    """
    API endpoint that allows activity to be viewed or edited.
    """
    queryset = RequestProviderResult.objects
    lookup_field = 'request_id'
    serializer_class = ProviderVerificationRequestResultSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    permission_classes = [
        permissions.ProviderPermission
    ]
    '''
    filterset_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    search_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    '''

    @action(detail=True, methods=['POST'])
    def status(self, request, *args, **kwargs):
        """
            Change enrolment sample validation status
        """
        status = request.data.get('status')
        if status is None:
            raise ValidationError('status field is required')
        try:
            status = int(status)
        except:
            raise ValidationError('status must be an integer')

        if status < 0 or status > 7:
            raise ValidationError('status value out of range')

        request_result = get_object_or_404(self.queryset,
                                           provider_id=kwargs['parent_lookup_provider_id'],
                                           request_id=kwargs['request_id'])

        if request_result.status == 1:
            raise ValidationError('Cannot change processed results status.')

        request_result.status = status
        request_result.save()

        return Response(self.serializer_class(instance=request_result).data)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for launcher token authentication """
import pytest


@pytest.mark.django_db
def test_api_authentication_launchertoken(rest_api_client, api_client, lapi_client, institution_course_test_case):

    # Generate a token for dashboards
    dash_token = api_client.get_launcher_token('DASHBOARD', institution_course_test_case['user'].institutionuser)

    assert dash_token is not None
    assert 'id' in dash_token
    assert 'token' in dash_token

    # Authenticate using the token
    resp_dash_token = rest_api_client.post('/api/v2/auth/token', data=dash_token)

    assert resp_dash_token is not None
    assert resp_dash_token.status_code == 200
    assert 'access_token' in resp_dash_token.data
    assert 'refresh_token' in resp_dash_token.data

    # Authenticate using the token
    validation_jwt_dash = api_client.validate_token(resp_dash_token.data['access_token'])

    assert validation_jwt_dash is not None
    assert validation_jwt_dash['pk'] == institution_course_test_case['user'].institutionuser.id

    # Generate a token for LAPI
    try:
        api_client.get_launcher_token('LAPI', institution_course_test_case['user'].institutionuser)
        pytest.fail('No learner users are able to request token for LAPI')
    except Exception:
        pass

    lapi_token = api_client.get_launcher_token('LAPI', institution_course_test_case['learner'].institutionuser)

    assert lapi_token is not None
    assert 'id' in lapi_token
    assert 'token' in lapi_token

    # Authenticate using the token
    resp_lapi_token = rest_api_client.post('/api/v2/auth/token', data=lapi_token)

    assert resp_lapi_token is not None
    assert resp_lapi_token.status_code == 200
    assert 'access_token' in resp_lapi_token.data
    assert 'refresh_token' in resp_lapi_token.data

    # Authenticate using the token for LAPI
    validation_jwt_lapi = lapi_client.validate_token(resp_lapi_token.data['access_token'])

    assert validation_jwt_lapi is not None
    assert validation_jwt_lapi['pk'] == institution_course_test_case['learner'].institutionuser.id

    # Authenticate using the token for API
    validation_jwt_api = api_client.validate_token(resp_lapi_token.data['access_token'])

    assert validation_jwt_api is not None
    assert validation_jwt_api['pk'] == institution_course_test_case['learner'].institutionuser.id

    # Request token for API
    try:
        api_client.get_launcher_token('API', institution_course_test_case['user'].institutionuser)
        pytest.fail('Launcher token for API should not be allowed')
    except Exception:
        pass



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" ActivityInstrument views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import VLECourseActivityInstrumentSerializer
from tesla_ce.models import ActivityInstrument


# pylint: disable=too-many-ancestors
class VLECourseActivityInstrumentViewSet(viewsets.ModelViewSet):
    """
        API endpoint that allows activity instruments to be viewed or edited.
    """
    serializer_class = VLECourseActivityInstrumentSerializer
    permission_classes = [
        permissions.VLEPermission
    ]
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    filterset_fields = ['instrument_id']
    search_fields = ['instrument_id']

    def get_queryset(self):
        queryset = ActivityInstrument.objects
        if self.kwargs['parent_lookup_activity_id']:
            queryset = queryset.filter(
                activity_id=self.kwargs['parent_lookup_activity_id']
            )
        return queryset.all().order_by('id')



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""UI Option serialize module."""
from rest_framework import serializers

from tesla_ce.models import UIOption


class InstitutionUIOptionSerializer(serializers.ModelSerializer):
    """UI Option serialize model module."""

    institution_id = serializers.HiddenField(default=None, allow_null=True)
    is_global = serializers.SerializerMethodField()
    roles = serializers.ReadOnlyField()

    class Meta:
        model = UIOption
        fields = ['id', 'route', 'enabled', 'roles', 'user', 'institution_id', 'is_global']

    def get_is_global(self, instance):
        return instance.institution is None and instance.user is None

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['institution_id'] = self.context['view'].kwargs['parent_lookup_institution_id']

        # Check that this route is not already defined for the institution and same user
        qs = UIOption.objects.filter(route=attrs['route'],
                                     institution__id=self.context['view'].kwargs['parent_lookup_institution_id'])
        if attrs['user'] is not None:
            qs =  qs.filter(user=attrs['user'])
        if self.instance is not None:
            qs = qs.exclude(id=self.instance.id)
        num_routes = qs.count()
        if num_routes > 0:
            raise serializers.ValidationError(detail="This route already exists")

        return super().validate(attrs)

    def validate_route(self, value):
        if value is None:
            raise serializers.ValidationError(detail="Route cannot be empty")

        if UIOption.objects.filter(route=value, institution=None, user=None).count() == 0:
            raise serializers.ValidationError(detail="Invalid Route")

        return value

    def validate_user(self, value):
        if value is not None and value.institution_id != int(
                self.context['view'].kwargs['parent_lookup_institution_id']
        ):
            raise serializers.ValidationError(detail="User must be from the institution")
        return value



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" InformedConsentDocument model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .informed_consent import InformedConsent


def get_upload_path(instance, filename):
    """
        Build the path where the informed consent document will be stored

        :param instance: Informed consent document
        :param filename: Name of the file
        :return: Path to store the file
    """
    return '{}/ic/{}/{}/{}'.format(
        instance.consent.institution.id,
        instance.consent.version,
        instance.language,
        filename
    )


class InformedConsentDocument(BaseModel):
    """ InformedConsentDocument model. """
    consent = models.ForeignKey(InformedConsent, null=False, blank=False,
                                on_delete=models.CASCADE)

    language = models.CharField(max_length=30, null=False, blank=False)

    html = models.TextField(null=True, blank=True,
                            help_text="HTML version of IC")

    pdf = models.FileField(null=True, blank=True, upload_to=get_upload_path,
                           help_text=_("PDF version of IC."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('consent', 'language'),)

    def __repr__(self):
        return "<InformedConsentDocument(consent_id='%d', language='%s')>" % (
            self.consent.id, self.language)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" CourseGroup api serialize module."""
from rest_framework import serializers
from rest_framework.exceptions import NotFound

from tesla_ce.models import Course
from tesla_ce.models import CourseGroup
from .institution import InstitutionSerializer
from ..vle import VLESerializer


class InstitutionCourseGroupsField(serializers.PrimaryKeyRelatedField):
    """ Related Institution field"""
    def get_queryset(self):
        """
            Return the queryset for this field
            :return: Queryset
        """
        queryset = super().get_queryset()
        if 'parent_lookup_institution_id' in self.context['view'].kwargs:
            queryset = queryset.filter(institution_id=self.context['view'].kwargs['parent_lookup_institution_id'])
        if self.root.instance is not None:
            queryset = queryset.exclude(id=self.root.instance.id)
        return queryset


class InstitutionCourseGroupSerializer(serializers.ModelSerializer):
    """CourseGroup serialize model module."""

    institution = InstitutionSerializer(read_only=True)
    institution_id = serializers.HiddenField(default=None, allow_null=True)
    parent = InstitutionCourseGroupsField(allow_null=True, default=None,
                                          queryset=CourseGroup.objects)

    class Meta:
        model = CourseGroup
        exclude = ["courses"]

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['institution_id'] = self.context['view'].kwargs['parent_lookup_institution_id']

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)


class InstitutionCourseGroupCourseSerializer(serializers.ModelSerializer):
    """Course serialize class."""

    id = serializers.IntegerField()
    vle_id = serializers.IntegerField(read_only=True)
    vle_course_id = serializers.CharField(read_only=True)
    code = serializers.CharField(read_only=True)
    description = serializers.CharField(read_only=True)
    start = serializers.DateTimeField(read_only=True)
    end = serializers.DateTimeField(read_only=True)

    class Meta:
        model = Course
        exclude = ["learners", "instructors", "vle"]

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['group_id'] = self.context['view'].kwargs['parent_lookup_id']

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)

    def create(self, validated_data):
        """
            Add the course to the group and return the related course

            :param validated_data: Valid data from request
            :return: Course
        """
        try:
            group = CourseGroup.objects.get(id=validated_data['group_id'])
        except CourseGroup.DoesNotExist:
            raise NotFound('Course Group not found')
        try:
            course = Course.objects.get(id=validated_data['id'])
        except Course.DoesNotExist:
            raise NotFound('Course not found')
        group.courses.add(course)

        return course



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE VLE actions for Use Case tests """
import mock

from django.utils import timezone

from tests import auth_utils
from tests.conftest import get_random_string

from ..learner import get_data_object_from_session


def vle_create_course(vle):
    """
        The VLE creates a new course
        :param vle: Vle credentials
        :return: New created course
    """
    # Authenticate using VLE credentials
    client, config = auth_utils.client_with_approle_credentials(vle['role_id'], vle['secret_id'])

    # Get the VLE ID from configuration
    vle_id = config['vle_id']

    # Create a new course
    course_code = get_random_string(10)
    course_data = {
        'code': course_code,
        'description': "PyTest test course {}".format(course_code),
        'vle_course_id': get_random_string(5),
        'start': timezone.now() - timezone.timedelta(days=1),
        'end': timezone.now() + timezone.timedelta(weeks=4)
    }
    course_create_resp = client.post(
        '/api/v2/vle/{}/course/'.format(vle_id),
        data=course_data, format='json'
    )
    assert course_create_resp.status_code == 201
    course = course_create_resp.data

    return course


def vle_enrol_users(vle, course):
    """
        The VLE enrolls course learners and instructors
        :param vle: Vle credentials
        :param course: Course object
        :return: A tuple (instructors, learners) with enrolled instructors and learners
    """
    # Authenticate using VLE credentials
    client, config = auth_utils.client_with_approle_credentials(vle['role_id'], vle['secret_id'])

    # Get the VLE ID from configuration
    vle_id = config['vle_id']

    # Create learners
    learners = []
    for _ in range(2):
        username = get_random_string(10)
        email = '{}@tesla-ce.eu'.format(username)
        learner_data = {
            'username': username,
            'uid': username,
            'email': email,
            'first_name': username[:5],
            'last_name': username[5:]
        }
        learner_create_resp = client.post(
            '/api/v2/vle/{}/course/{}/learner/'.format(vle_id, course['id']),
            data=learner_data,
            format='json'
        )
        assert learner_create_resp.status_code == 201
        learners.append(learner_create_resp.data)

    # Create instructors
    instructors = []
    for _ in range(2):
        username = get_random_string(10)
        email = '{}@tesla-ce.eu'.format(username)
        learner_data = {
            'username': username,
            'uid': username,
            'email': email,
            'first_name': username[:5],
            'last_name': username[5:]
        }
        instructor_create_resp = client.post(
            '/api/v2/vle/{}/course/{}/instructor/'.format(vle_id, course['id']),
            data=learner_data,
            format='json'
        )
        assert instructor_create_resp.status_code == 201
        instructors.append(instructor_create_resp.data)

    return instructors, learners


def vle_create_activity(vle, course):
    """
        The VLE creates a new activity
        :param vle: Vle credentials
        :param course: Course object
        :return: New created activity
    """
    # Authenticate using VLE credentials
    client, config = auth_utils.client_with_approle_credentials(vle['role_id'], vle['secret_id'])

    # Get the VLE ID from configuration
    vle_id = config['vle_id']

    # Create an activity
    activity_create_resp = client.post(
        '/api/v2/vle/{}/course/{}/activity/'.format(vle_id, course['id']),
        data={
            'enabled': True,
            'code': get_random_string(10),
            'description': "PyTest test course",
            'vle_course_id': get_random_string(5),
            'vle_activity_type': 'quiz',
            'vle_activity_id': get_random_string(3)
        },
        format='json'
    )
    assert activity_create_resp.status_code == 201

    return activity_create_resp.data


def vle_check_learner_ic(vle, course, learner, missing=True):
    """
        VLE check the status of the Informed Consent of the learner
        :param vle: VLE credentials
        :param course: Course object
        :param learner: Learner object
        :param missing: True if it is expected that IC is not still accepted or False otherwise
    """
    # Authenticate using VLE credentials
    client, config = auth_utils.client_with_approle_credentials(vle['role_id'], vle['secret_id'])

    # Get the VLE ID from configuration
    vle_id = config['vle_id']

    # Create an activity
    learner_data_resp = client.get(
        '/api/v2/vle/{}/course/{}/learner/{}/'.format(vle_id, course['id'], learner['id']),
    )
    assert learner_data_resp.status_code == 200
    if missing:
        assert learner_data_resp.data['ic_status'].startswith('NOT_VALID')
    else:
        assert learner_data_resp.data['ic_status'].startswith('VALID')


def vle_create_assessment_session(vle, learner, activity, ic=True, enrolment=True, options=None):
    """
        The VLE creates an assessment session.
        :param vle: VLE object
        :param learner: Learner object
        :param activity: Activity object
        :param ic: True if IC is expected to be accepted or False otherwise
        :param enrolment: True if enrolment is expected to be performed or False otherwise
        :param options: Dictionary of options to customize the client
        :return: List of missing instruments
    """
    # Authenticate using VLE credentials
    client, config = auth_utils.client_with_approle_credentials(vle['role_id'], vle['secret_id'])

    # Get the VLE ID from configuration
    vle_id = config['vle_id']

    # Get an evaluation session
    create_session_resp = client.post(
        '/api/v2/vle/{}/assessment/'.format(vle_id),
        data={
            'vle_activity_type': activity['vle_activity_type'],
            'vle_activity_id': activity['vle_activity_id'],
            'vle_learner_uid': learner['uid'],
            'options': options
        },
        format='json'
    )
    if not ic:
        assert create_session_resp.status_code == 406
        assert create_session_resp.data['status'] == 2  # MISSING IC
        assert 'INFORMED CONSENT' in create_session_resp.data['message'].upper()
    elif not enrolment:
        assert create_session_resp.status_code == 406
        assert create_session_resp.data['status'] == 4  # MISSING ENROLMENT
        assert 'ENROLMENT' in create_session_resp.data['message'].upper()
    else:
        assert create_session_resp.status_code == 200

    # Check missing instruments
    return create_session_resp.data


def vle_create_launcher(vle, user, session=None):
    """
        The VLE creates a launcher for the user
        :param vle: VLE object
        :param user: User object
        :param session: Assessment session object
        :return: New launcher data
    """
    # Authenticate using VLE credentials
    client, config = auth_utils.client_with_approle_credentials(vle['role_id'], vle['secret_id'])

    # Get the VLE ID from configuration
    vle_id = config['vle_id']

    # Get data for launcher
    data = {
        'vle_user_uid': user['uid']
    }

    if session is not None:
        data['session_id'] = session['id']

    # Create a launcher
    launcher_create_resp = client.post('/api/v2/vle/{}/launcher/'.format(vle_id), data=data)
    assert launcher_create_resp.status_code == 200

    launcher = launcher_create_resp.data
    return launcher


def vle_send_activity(vle, assessment_session, document, omit_session=False):
    """
        The VLE send the submission performed by learner to TeSLA if required
        :param vle: VLE object
        :param assessment_session: Assessment session object
        :param document: Submitted document
        :param omit_session: If True, activity is sent without session
        :return: Pending tasks
    """
    # List simulating the storage queue
    pending_tasks_storage = []
    # List simulating the verification queue
    pending_tasks_verification = []
    # Object simulating the providers verification queues
    pending_provider_tasks_verification = []

    def create_request_test(*args, **kwargs):
        pending_tasks_storage.append(('create_request', args, kwargs))

    def verify_request_test(*args, **kwargs):
        pending_tasks_verification.append(('verify_request', args, kwargs))

    def verify_request_prov_test(*args, **kwargs):
        pending_provider_tasks_verification.append(('verify_request', args, kwargs))

    # Authenticate using VLE credentials
    client, config = auth_utils.client_with_approle_credentials(vle['role_id'], vle['secret_id'])

    # Get the options for verification
    session_data = get_data_object_from_session(assessment_session)

    # If no document is required just return an empty list of tasks
    if document is None:
        assert 'activity' not in session_data['sensors']
        return []

    # Get required data
    vle_id = config['vle_id']
    learner_id = session_data['learner']['learner_id']
    course_id = session_data['activity']['course']['id']
    activity_id = session_data['activity']['id']

    # Get the list of instruments waiting for document
    activity_instruments_resp = client.get('/api/v2/vle/{}/course/{}/activity/{}/attachment/{}/'.format(
        vle_id, course_id, activity_id, learner_id
    ))
    assert activity_instruments_resp.status_code == 200
    instruments = [inst['instrument'] for inst in activity_instruments_resp.data]

    # Create the document data
    document_data = {
                'session_id': session_data['session_id'],
                'instruments': instruments,
                'metadata': {
                    'mimetype': document['mimetype'],
                    'filename': document['filename'],
                    'created_at': timezone.now(),
                    'context': {}
                },
                'data': document['content']
            }
    if omit_session:
        del document_data['session_id']

    # Make a submission of the document
    with mock.patch('tesla_ce.tasks.requests.verification.create_request.apply_async', create_request_test):
        data_sent_resp = client.post(
            '/api/v2/vle/{}/course/{}/activity/{}/attachment/{}/'.format(
                vle_id, course_id, activity_id, learner_id
            ),
            data=document_data,
            format='json'
        )
        assert data_sent_resp.status_code == 200
        assert data_sent_resp.data['status'] == 'OK'

    # Run storage tasks
    with mock.patch('tesla_ce.tasks.requests.verification.verify_request.apply_async', verify_request_test):
        from tesla_ce.tasks.requests.verification import create_request
        for task in pending_tasks_storage:
            assert task[0] == 'create_request'
            create_request(**task[2]['kwargs'])

    # Run verification tasks
    with mock.patch('tesla_ce.tasks.requests.verification.verify_request.apply_async', verify_request_prov_test):
        from tesla_ce.tasks.requests.verification import verify_request
        for task in pending_tasks_verification:
            assert task[0] == 'verify_request'
            verify_request(task[1][0][0])

    return pending_provider_tasks_verification


def vle_activity_report(vle, activity):
    """
        VLE request the results for the activity
        :param vle: VLE object
        :param activity: Activity object
    """
    # Authenticate using VLE credentials
    client, config = auth_utils.client_with_approle_credentials(vle['role_id'], vle['secret_id'])

    # Get required data
    vle_id = config['vle_id']

    # Get required data
    course_id = activity['course']['id']
    activity_id = activity['id']

    # Get the list of reports from the activity
    activity_reports_resp = client.get('/api/v2/vle/{}/course/{}/activity/{}/report/'.format(
        vle_id,
        course_id,
        activity_id
    ))
    assert activity_reports_resp.status_code == 200
    reports = activity_reports_resp.data['results']

    # Get the detail for each report
    for report in reports:
        reports_detail_resp = client.get('/api/v2/vle/{}/course/{}/activity/{}/report/{}/'.format(
            vle_id,
            course_id,
            activity_id,
            report['id']
        ))
        assert reports_detail_resp.status_code == 200
        report['detailed_report'] = reports_detail_resp.data

        request_list = []
        final = False
        while not final:
            report_requests = client.get(
                '/api/v2/vle/{}/course/{}/activity/{}/learner/{}/request/?offset={}'.format(
                    vle_id,
                    course_id,
                    activity_id,
                    report['learner']['id'],
                    len(request_list)
            ))
            assert report_requests.status_code == 200
            request_list += report_requests.data['results']
            if len(request_list) >= report_requests.data['count']:
                final = True
        report['detailed_report']['requests'] = request_list

    return reports


def vle_close_assessment_session(vle, assessment_session):
    """
        The VLE creates an assessment session.
        :param vle: VLE object
        :param assessment_session: The assessment session to close
    """
    # Authenticate using VLE credentials
    client, config = auth_utils.client_with_approle_credentials(vle['role_id'], vle['secret_id'])

    # Get the VLE ID from configuration
    vle_id = config['vle_id']

    # Get an evaluation session
    close_session_resp = client.post(
        '/api/v2/vle/{}/assessment/'.format(vle_id),
        data={
            'session_id': assessment_session['id'],
            'close': True
        },
        format='json'
    )
    assert close_session_resp.status_code == 200
    assert close_session_resp.data['closed_at'] is not None

    # Check missing instruments
    return close_session_resp.data



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Request Provider Result model module."""

from django.db import models
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .histogram_activity_instrument import HistogramActivityInstrument
from .histogram_activity_provider import HistogramActivityProvider
from .histogram_learner_instrument import HistogramLearnerInstrument
from .histogram_learner_provider import HistogramLearnerProvider
from .message import Message
from .provider import Provider
from .request import Request
from .request_result import RESULT_CODE
from .request_result import RESULT_STATUS


class RequestProviderResult(BaseModel):
    """ Request result for a provider. """

    request = models.ForeignKey(Request, null=False,
                                on_delete=models.CASCADE, help_text=_('Request related to this result'))

    provider = models.ForeignKey(Provider, null=False,
                                 on_delete=models.CASCADE, help_text=_('Provider related to this result'))

    status = models.IntegerField(choices=RESULT_STATUS, null=False,
                                 default=0,
                                 help_text=_('Status for this result'))

    result = models.FloatField(null=True, default=0.0,
                               help_text=_('Normalized result value'))

    error_message = models.TextField(null=True,
                                     help_text=_("Error message when status is error"))

    message_code = models.ForeignKey(Message, null=True, default=None,
                                     on_delete=models.SET_NULL,
                                     help_text=_("Related message code"))

    code = models.IntegerField(choices=RESULT_CODE, null=False, default=0,
                               help_text=_("Result code provided after performing the verification process"))

    audit = models.FileField(max_length=250, null=False, blank=False,
                             help_text=_("Audit data for this result."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('provider', 'request'),)

    def __repr__(self):
        return "<RequestProviderResult(request='%s', provider='%s', status='%r', " \
               "result='%r', code='%r')>" % (
                   self.request, self.provider.acronym, self.get_status_display(),
                   self.result, self.get_code_display()
        )

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        super().save(force_insert, force_update, using, update_fields)

        # Update histograms
        if self.status == 1:
            bin = 'b{}'.format(min(int(self.result * 10), 9))
            with transaction.atomic():
                try:
                    h1 = HistogramLearnerInstrument.objects.get(learner=self.request.learner,
                                                                instrument=self.provider.instrument)
                    setattr(h1, bin, getattr(h1, bin) + 1)
                    h1.save()
                except HistogramLearnerInstrument.DoesNotExist:
                    h1 = HistogramLearnerInstrument(learner=self.request.learner,
                                                    instrument=self.provider.instrument)
                    setattr(h1, bin, 1)
                    h1.save()
                try:
                    h2 = HistogramLearnerProvider.objects.get(learner=self.request.learner,
                                                              provider=self.provider)
                    setattr(h2, bin, getattr(h2, bin) + 1)
                    h2.save()
                except HistogramLearnerProvider.DoesNotExist:
                    h2 = HistogramLearnerProvider(learner=self.request.learner,
                                                  provider=self.provider)
                    setattr(h2, bin, 1)
                    h2.save()
                try:
                    h3 = HistogramActivityInstrument.objects.get(activity=self.request.activity,
                                                                 instrument=self.provider.instrument)
                    setattr(h3, bin, getattr(h3, bin) + 1)
                    h3.save()
                except HistogramActivityInstrument.DoesNotExist:
                    h3 = HistogramActivityInstrument(activity=self.request.activity,
                                                     instrument=self.provider.instrument)
                    setattr(h3, bin, 1)
                    h3.save()
                try:
                    h4 = HistogramActivityProvider.objects.get(activity=self.request.activity,
                                                               provider=self.provider)
                    setattr(h4, bin, getattr(h4, bin) + 1)
                    h4.save()
                except HistogramActivityProvider.DoesNotExist:
                    h4 = HistogramActivityProvider(activity=self.request.activity,
                                                   provider=self.provider)
                    setattr(h4, bin, 1)
                    h4.save()



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Module with authentication utils for tests """
from rest_framework.test import APIClient


def client_with_user_obj(user):
    """
        Get an API client instance authenticated with provided user
        :param user: User object
        :return: API Client instance
    """
    client = APIClient()
    client.force_authenticate(user=user)

    return client


def get_profile(client):
    """
        Get the profile of authenticated user
        :param client: API Client with user authentication
        :return: Profile data
    """
    # Get user profile
    profile_resp = client.get('/api/v2/auth/profile')
    assert profile_resp.status_code == 200

    return profile_resp.data


def client_with_token_credentials(access_token, refresh_token=None):
    """
        Get an API client instance authenticated with token object
        :param access_token: Token for authentication
        :param refresh_token: Token for token renewal
        :return: Client authenticated for the user
    """
    client = APIClient()

    # Assign credentials
    client.credentials(HTTP_AUTHORIZATION='JWT {}'.format(access_token))

    return client


def client_with_user_credentials(email, password):
    """
        Get an API client instance authenticated with provided user credentials
        :param email: User's email
        :param password: User's password
        :return: API Client instance
    """
    client = APIClient()

    # TODO: Remove fake authentication
    password = email

    auth_resp = client.post('/api/v2/auth/login', data={
        'email': email,
        'password': password
    })
    assert auth_resp.status_code == 200
    assert 'token' in auth_resp.data

    # Get client
    return client_with_token_credentials(
        auth_resp.data['token']['access_token'],
        auth_resp.data['token']['refresh_token']
    )


def client_with_approle_credentials(role_id, secret_id):
    """
        Get an API client instance authenticated with provided AppRole module credentials
        :param role_id: The module's RoleID
        :param secret_id: The module's SecretID
        :return: A tuple (client, configuration) with the client instance and module configuration
    """
    client = APIClient()

    auth_resp = client.post('/api/v2/auth/approle', data={
        'role_id': role_id,
        'secret_id': secret_id
    })
    assert auth_resp.status_code == 200
    assert 'token' in auth_resp.data

    # Assign credentials
    token = auth_resp.data['token']['access_token']
    client.credentials(HTTP_AUTHORIZATION='JWT {}'.format(token))

    return client, auth_resp.data


def client_with_launcher_credentials(launcher):
    """
        Get an API client instance authenticated with provided launcher
        :param launcher: Launcher object
        :return: User credentials from launcher
    """
    client = APIClient()

    auth_resp = client.post('/api/v2/auth/token',
        data={'id': launcher['id'], 'token': launcher['token']},
        format='json'
    )
    assert auth_resp.status_code == 200
    assert 'access_token' in auth_resp.data
    assert 'refresh_token' in auth_resp.data

    # Get client
    return client_with_token_credentials(
        auth_resp.data['access_token'],
        auth_resp.data['refresh_token']
    )


def refresh_token(access_token, refresh_token):
    """
        Refresh a token
        :param access_token: Access JWT token
        :param refresh_token: Refresh JWT token
        :return: A new token pair
    """
    client = APIClient()

    # Assign credentials
    client.credentials(HTTP_AUTHORIZATION='JWT {}'.format(refresh_token))

    # Refresh token
    refresh_resp = client.post('/api/v2/auth/token/refresh',
                               data={'token': access_token},
                               format='json')
    assert refresh_resp.status_code == 200
    assert 'token' in refresh_resp.data
    assert 'access_token' in refresh_resp.data['token']
    assert 'refresh_token' in refresh_resp.data['token']

    return refresh_resp.data['token']



# Generated by Django 3.0 on 2020-09-30 16:42

import django.utils.timezone
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0010_auto_20200930_1829'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='instructor',
            options={},
        ),
        migrations.AlterModelOptions(
            name='learner',
            options={},
        ),
        migrations.AddField(
            model_name='institutionuser',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, help_text='Date when user was created'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='institutionuser',
            name='locale',
            field=models.CharField(default=None, help_text='Default locale for this user', max_length=10, null=True),
        ),
        migrations.AddField(
            model_name='institutionuser',
            name='updated_at',
            field=models.DateTimeField(auto_now=True, help_text='Last user modification'),
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
import pytest

from tesla_ce.lib import TeslaVaultException


@pytest.mark.django_db
def test_create_learner_token(api_client, lapi_client, institution_course_test_case):
    assert api_client is not None
    assert lapi_client is not None

    learner = institution_course_test_case['learner']
    if learner is None:
        pytest.skip('No learners in DB')

    # Get learner object
    learner = learner.institutionuser.learner

    # Generate the token in the API
    token = api_client.get_learner_token_pair(learner, scope='/api/*')
    assert token is not None
    assert 'access_token' in token
    assert 'refresh_token' in token

    # Check that this token is valid for API
    validation_api = api_client.validate_token(token['access_token'])
    assert validation_api is not None
    assert isinstance(validation_api, dict)
    assert validation_api['group'] == 'learners'

    # Check that this token is valid for Learners API
    validation_lapi = lapi_client.validate_token(token['access_token'])
    assert validation_lapi is not None
    assert isinstance(validation_lapi, dict)
    assert validation_lapi['group'] == 'learners'

    # Ensure that lapi cannot create learner tokens
    try:
        lapi_client.get_learner_token_pair(learner, scope='/api/*')
        pytest.fail('LAPI Module is able to generate learner tokens')
    except TeslaVaultException:
        pass


@pytest.mark.django_db
def test_create_instructor_token(api_client, lapi_client, institution_course_test_case):
    assert api_client is not None
    assert lapi_client is not None

    instructor = institution_course_test_case['instructor']
    if instructor is None:
        pytest.skip('No instructors in DB')

    # Get instructor object
    instructor = instructor.institutionuser.instructor
    scope = {
        'activity': 'my_activity'
    }

    # Generate the token in the API
    token = api_client.get_instructor_token(instructor.email, scope)
    assert token is not None

    # Check that this token is valid for API
    validation_api = api_client.validate_token(token)
    assert validation_api is not None
    assert isinstance(validation_api, dict)
    assert validation_api['group'] == 'instructors'

    # Ensure that lapi cannot create instructor tokens
    try:
        lapi_client.get_instructor_token(instructor.email, scope)
        pytest.fail('LAPI Module is able to generate instructor tokens')
    except TeslaVaultException:
        pass


@pytest.mark.django_db
def test_create_user_token(api_client, lapi_client, institution_test_case):
    assert api_client is not None
    assert lapi_client is not None

    user = institution_test_case['user'].institutionuser
    if user is None:
        pytest.skip('No institution users in DB')

    scope = ['/api/*']

    # Generate the token in the API
    token = api_client.get_user_token_pair(user=user, scope=scope)
    assert token is not None
    assert 'access_token' in token
    assert 'refresh_token' in token

    # Check that this token is valid for API
    validation_api = api_client.validate_token(token['access_token'])
    assert validation_api is not None
    assert isinstance(validation_api, dict)
    assert validation_api['group'] == 'users'

    # Ensure that lapi cannot create users tokens
    try:
        lapi_client.get_user_token_pair(user=user, scope=scope)
        pytest.fail('LAPI Module is able to generate user tokens')
    except TeslaVaultException:
        pass


@pytest.mark.django_db
def test_create_module_token(api_client, lapi_client):
    assert api_client is not None
    assert lapi_client is not None

    scope = {
    }

    # Generate the token in the API
    api_token = api_client.get_module_token(scope)
    assert api_token is not None

    # Check that this token is valid for API
    validation_api = api_client.validate_token(api_token)
    assert validation_api is not None
    assert isinstance(validation_api, dict)
    assert validation_api['group'].startswith('module_')
    assert validation_api['sub'] == 'api'

    # Check that other modules cannot generate tokens
    try:
        lapi_client.get_module_token(scope)
        pytest.fail('Dashboards Module is able to generate module tokens')
    except TeslaVaultException:
        pass

    # Generate the token for LAPI
    lapi_token = api_client.get_module_token(scope, module_id='lapi')

    # Check that this token is valid for API
    validation_api = api_client.validate_token(lapi_token)
    assert validation_api is not None
    assert isinstance(validation_api, dict)
    assert validation_api['group'].startswith('module_')
    assert validation_api['sub'] == 'lapi'

    # Check that this token is valid for LAPI
    validation_lapi = lapi_client.validate_token(lapi_token)
    assert validation_lapi is not None
    assert isinstance(validation_lapi, dict)
    assert validation_lapi['group'].startswith('module_')
    assert validation_lapi['sub'] == 'lapi'



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Webhooks views module """
from rest_framework.views import APIView
from rest_framework.views import Response
from rest_framework.views import status

from tesla_ce.apps.webhooks import exceptions
from .client.base import BaseWebhookMessage


class WebhooksView(APIView):
    """
        Receiver for webhooks
    """
    def post(self, request):
        """
            Webhook process

            :param request:
            :return:
        """
        try:
            msg = BaseWebhookMessage.get_webhook_object(request)
            msg.save()
            msg.process()
        except exceptions.WebhookAuthException as wae:
            return Response(status=status.HTTP_403_FORBIDDEN, data={'message': wae.__str__()})
        except exceptions.WebhookClientNotFoundException as wcnfe:
            return Response(status=status.HTTP_404_NOT_FOUND, data={'message': wcnfe.__str__()})
        except exceptions.WebhookException as we:
            return Response(status=status.HTTP_400_BAD_REQUEST, data={'message': we.__str__()})

        return Response(status=status.HTTP_200_OK)


class WebhooksEditView(APIView):
    """
        Webhooks edit view
    """

    def get(self, request, format=None, webhook_id=None):
        """
            Get webhook data

            :param request:
            :param webhook_id: Unique identifier for webhook
            :return:
        """
        msg_data = None
        try:
            msg = BaseWebhookMessage.load(webhook_id, request)
            msg_data = msg.get_data()
        except exceptions.WebhookAuthException as wae:
            return Response(status=status.HTTP_403_FORBIDDEN, data={'message': wae.__str__()})
        except exceptions.WebhookClientNotFoundException as wcnfe:
            return Response(status=status.HTTP_404_NOT_FOUND, data={'message': wcnfe.__str__()})
        except exceptions.WebhookException as we:
            return Response(status=status.HTTP_400_BAD_REQUEST, data={'message': we.__str__()})

        return Response(status=status.HTTP_200_OK, data=msg_data)

    def post(self, request, webhook_id=None):
        """
            Update webhook

            :param request:
            :param webhook_id: Unique identifier for webhook
            :return:
        """
        try:
            msg = BaseWebhookMessage.load(webhook_id, request)
            message = None
            status_val = None
            if "status" in request.data:
                status_val = request.data['status']
            if "message" in request.data:
                message = request.data['message']
            msg.update_status(status_val, message)
        except exceptions.WebhookAuthException as wae:
            return Response(status=status.HTTP_403_FORBIDDEN, data={'message': wae.__str__()})
        except exceptions.WebhookClientNotFoundException as wcnfe:
            return Response(status=status.HTTP_404_NOT_FOUND, data={'message': wcnfe.__str__()})
        except exceptions.WebhookException as we:
            return Response(status=status.HTTP_400_BAD_REQUEST, data={'message': we.__str__()})

        return Response(status=status.HTTP_200_OK)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango command to generate deployment scripts for a VLE """
from ..base import TeslaDeployCommand


class Command(TeslaDeployCommand):
    """ Command to generate VLE deployment scripts """
    help = 'Generate VLE deployment scripts'
    requires_system_checks = '__all__'

    def add_arguments(self, parser):
        """
            Define custom arguments for this command

            :param parser: Input command parser instance
        """
        # Set default arguments
        super().add_arguments(parser)

        parser.add_argument(
            '--client-id',
            help='Client ID for LTI 1.3',
            default=None,
        )
        parser.add_argument(
            '--type',
            choices=['moodle'],
            default='moodle',
            help='Type of VLE',
        )

    def custom_handle(self):
        """
            Custom actions for this command
        """
        # Export the deployment scripts
        self.client.export_vle_scripts(
            type=self._options['type'],
            client_id=self._options['client_id'],
            output=self._options['out'],
            mode=self._options['mode'])
        self.stdout.write(self.style.SUCCESS('Deployment scripts written at {}'.format(self._options['out'])))



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" API Views for TeSLA Identity Provieder (TIP) """
from rest_framework.views import APIView
from rest_framework.views import Response
from rest_framework.views import status

from tesla_ce.models import Institution
from tesla_ce.models import Learner
from ..serializers.tip import LearnerIdSerializer
from ..serializers.tip import LearnerSerializer
from ..serializers.tip import MultipleLearnerIdSerializer


class LearnerId(APIView):
    """
        Get the Learner TeSLA Id from mail
    """
    def post(self, request, format=None):

        serializer = LearnerIdSerializer(data=request.data, context={'request': request})

        if serializer.is_valid():
            # Use default institution
            try:
                institution = Institution.objects.get(pk=1)
            except Institution.DoesNotExist:
                return Response(status=status.HTTP_404_NOT_FOUND)
            # Search for the learner
            try:
                learner = Learner.objects.get(institution=institution, email=serializer.validated_data['mail'])
            except Learner.DoesNotExist:
                learner = Learner.objects.create(uid=serializer.validated_data['mail'],
                                                 email=serializer.validated_data['mail'],
                                                 username=serializer.validated_data['mail'],
                                                 institution=institution)
            return Response(LearnerSerializer(learner).data)

        return Response(status=status.HTTP_400_BAD_REQUEST, data={'errors': serializer.errors})


class MultipleLearnerId(APIView):
    """
        Get the Learner TeSLA Ids from a list of mails
    """
    def post(self, request, format=None):
        serializer = MultipleLearnerIdSerializer(data=request.data, context={'request': request})

        if serializer.is_valid():
            # Use default institution
            try:
                institution = Institution.objects.get(pk=1)
            except Institution.DoesNotExist:
                return Response(status=status.HTTP_404_NOT_FOUND)
            # Search for the learners
            learners = Learner.objects.filter(institution=institution,
                                              email__in=serializer.validated_data['mails']
                                              ).all()
            return Response(LearnerSerializer(learners, many=True).data)

        return Response(status=status.HTTP_400_BAD_REQUEST, data={'errors': serializer.errors})



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""VLE api serialize module."""
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from tesla_ce.apps.api.utils import JSONField

from tesla_ce.models import AssessmentSession
from tesla_ce.models import AssessmentSessionData


class VLENewAssessmentSessionBodySerializer(serializers.Serializer):
    vle_activity_type = serializers.CharField(required=False, allow_blank=False, allow_null=False)
    vle_activity_id = serializers.CharField(required=False, allow_blank=False, allow_null=False)
    vle_learner_uid = serializers.CharField(required=False, allow_blank=False, allow_null=False)
    locale = serializers.CharField(required=False, allow_null=True, default=None)
    max_ttl = serializers.IntegerField(required=False, allow_null=False, default=120)
    session_id = serializers.IntegerField(required=False, allow_null=True, default=None)
    redirect_reject_url = serializers.URLField(required=False, allow_null=True, default=None)
    reject_message = serializers.CharField(required=False, allow_null=True, default=None)
    options = JSONField(required=False, allow_null=True, default=True)
    close = serializers.BooleanField(required=False, allow_null=True, default=False)

    class Meta:
        fields = ["vle_activity_type", "vle_activity_id", "vle_learner_uid", "locale", "max_ttl",
                  "session_id", "redirect_reject_url", "reject_message", "options", "close"]

    def validate(self, attrs):
        """
            Vaidate data
            :param attrs: Provided data
            :return: validated data
        """
        if 'session_id' not in attrs or attrs['session_id'] is None:
            if 'vle_activity_type' not in attrs or attrs['vle_activity_type'] is None:
                raise ValidationError('Missing vle_activity_type value')
            if 'vle_activity_id' not in attrs or attrs['vle_activity_id'] is None:
                raise ValidationError('Missing vle_activity_id value')
            if 'vle_learner_uid' not in attrs or attrs['vle_learner_uid'] is None:
                raise ValidationError('Missing vle_learner_uid value')
        return super().validate(attrs)


class VLENewAssessmentSessionDataSerializer(serializers.ModelSerializer):
    """New assessment session data serialize module."""

    class Meta:
        model = AssessmentSessionData
        fields = "__all__"


class VLENewAssessmentSessionSerializer(serializers.ModelSerializer):
    """New assessment session serialize module."""
    data = VLENewAssessmentSessionDataSerializer(source='assessmentsessiondata', many=False, read_only=True)

    class Meta:
        model = AssessmentSession
        fields = "__all__"



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Enrolment model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .enrolment_sample import EnrolmentSample
from .learner import Learner
from .learner import get_learner_enrolment
from .provider import Provider


def get_upload_path(instance, filename):
    """
        Build the path where the informed consent document will be stored

        :param instance: Informed consent document
        :type instance: Enrolment
        :param filename: Name of the file
        :return: Path to store the file
    """
    return '{}/models/{}/{}/{}.json'.format(
        instance.learner.institution.id,
        instance.learner.learner_id,
        instance.provider.instrument.id,
        instance.provider.id
    )


class Enrolment(BaseModel):
    """ Enrollment model. """
    learner = models.ForeignKey(Learner, null=False,
                                on_delete=models.CASCADE,
                                help_text=_("Related learner"))

    provider = models.ForeignKey(Provider, null=False,
                                 on_delete=models.CASCADE,
                                 help_text=_("Related provider"))

    percentage = models.DecimalField(null=False, decimal_places=2,
                                     max_digits=5,
                                     help_text=_("Percentage of enrolment for this learner"))

    can_analyse = models.BooleanField(null=False, default=False,
                                      help_text=_("This provider can verify learner samples"))

    locked_at = models.DateTimeField(null=True, default=None,
                                     help_text=_("This model is locked for an update process"))

    locked_by = models.UUIDField(null=True, default=None, blank=None,
                                 help_text=_("Task blocking this model"))

    model = models.FileField(max_length=250, null=True, default=None, upload_to=get_upload_path,
                             help_text=_('Path to stored model'))

    model_samples = models.ManyToManyField(EnrolmentSample, related_name='models',
                                           help_text=_('Enrolment samples used to create this model'))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('learner', 'provider'),)

    @property
    def is_locked(self):
        return self.locked_at is not None and self.locked_by is not None

    @property
    def model_total_samples(self):
        return self.model_samples.count()

    def __repr__(self):
        return "<Enrollment(learner='%s', provider='%r', percentage='%r'," \
               "can_analyse='%s', locked='%r', num_samples='%r')>" % (self.learner, self.provider, self.percentage,
                                                                      self.can_analyse, self.is_locked,
                                                                      self.model_total_samples)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        super().save(force_insert, force_update, using, update_fields)
        # Invalidate cached value
        get_learner_enrolment.invalidate(self.learner_id)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" VLE views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework.views import Response
from rest_framework.views import status
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce import get_default_client
from tesla_ce.apps.api import status as tesla_status
from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import VLELauncherBodySerializer
from tesla_ce.apps.api.v2.serializers import VLELauncherDataSerializer
from tesla_ce.apps.api.v2.serializers import VLENewAssessmentSessionBodySerializer
from tesla_ce.apps.api.v2.serializers import VLENewAssessmentSessionSerializer
from tesla_ce.apps.api.v2.serializers import VLESerializer
from tesla_ce.lib.exception import TeslaInvalidICException
from tesla_ce.lib.exception import TeslaMissingEnrolmentException
from tesla_ce.lib.exception import TeslaMissingICException
from tesla_ce.models import Activity
from tesla_ce.models import AssessmentSession
from tesla_ce.models import InstitutionUser
from tesla_ce.models import Learner
from tesla_ce.models import VLE


# pylint: disable=too-many-ancestors
class VLEViewSet(NestedViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    API endpoint that allows access to VLE related models.
    """
    queryset = VLE.objects.all().order_by('created_at')
    serializer_class = VLESerializer
    permission_classes = [
        permissions.VLEPermission
    ]
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    '''
    filterset_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    search_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    '''

    @action(detail=True, methods=['POST',],
            serializer_class=VLENewAssessmentSessionSerializer)
    def assessment(self, request, pk):
        """
            Create a new assessment session
        """
        serializer = VLENewAssessmentSessionBodySerializer(data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        try:
            if serializer.data['session_id'] is not None:
                assessment_session = AssessmentSession.objects.get(id=serializer.data['session_id'])

                # Close the session
                if serializer.data['close'] is not None and serializer.data['close']:
                    if not assessment_session.is_active:
                        raise ValidationError('Cannot close a closed session')
                    assessment_session.close(auto=False, close_related=True)
                elif not assessment_session.is_active:
                    raise ValidationError('Closed session')

                resp = self.serializer_class(assessment_session)
                return Response(resp.data)
            vle = VLE.objects.get(pk=pk)
            activity = Activity.objects.get(vle=vle, vle_activity_type=serializer.data['vle_activity_type'],
                                            vle_activity_id=serializer.data['vle_activity_id'])
            learner = Learner.objects.get(institution=vle.institution, uid=serializer.data['vle_learner_uid'])

        except VLE.DoesNotExist:
            return Response("VLE not found", status=status.HTTP_404_NOT_FOUND)
        except Activity.DoesNotExist:
            return Response("Activity not found", status=status.HTTP_404_NOT_FOUND)
        except Learner.DoesNotExist:
            return Response("Learner not found", status=status.HTTP_404_NOT_FOUND)
        except AssessmentSession.DoesNotExist:
            return Response("Session not found", status=status.HTTP_404_NOT_FOUND)

        # Create the assessment session
        try:
            assessment_session = get_default_client(
            ).create_assessment_session(activity=activity,
                                        learner=learner,
                                        locale=serializer.data['locale'],
                                        max_ttl=serializer.data['max_ttl'],
                                        redirect_reject_url=serializer.data['redirect_reject_url'],
                                        reject_message=serializer.data['reject_message'],
                                        options=serializer.data['options']
                                        )
        except TeslaMissingICException:
            return Response({
                'status': tesla_status.TESLA_MISSING_IC,
                'message': 'Missing Informed Consent'
            }, status=status.HTTP_406_NOT_ACCEPTABLE)
        except TeslaInvalidICException:
            return Response({
                'status': tesla_status.TESLA_INVALID_IC,
                'message': 'Invalid Informed Consent'
            }, status=status.HTTP_406_NOT_ACCEPTABLE)
        except TeslaMissingEnrolmentException as exc:
            return Response({
                'status': tesla_status.TESLA_MISSING_ENROLMENT,
                'message': 'Missing Enrolment',
                'enrolments': exc.args[0]
            }, status=status.HTTP_406_NOT_ACCEPTABLE)

        resp = self.serializer_class(assessment_session)
        return Response(resp.data)

    @action(detail=True, methods=['POST', ],
            serializer_class=VLELauncherDataSerializer)
    def launcher(self, request, pk):
        """
            Create a new launcher url for VLE user
        """
        serializer = VLELauncherBodySerializer(data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        try:
            vle = VLE.objects.get(pk=pk)
            user = InstitutionUser.objects.get(institution=vle.institution, uid=serializer.data['vle_user_uid'])
            session = None
            if serializer.data['session_id'] is not None:
                session = AssessmentSession.objects.get(id=serializer.data['session_id'])
        except VLE.DoesNotExist:
            return Response("VLE not found", status=status.HTTP_404_NOT_FOUND)
        except InstitutionUser.DoesNotExist:
            return Response("User not found", status=status.HTTP_404_NOT_FOUND)
        except AssessmentSession.DoesNotExist:
            return Response("Session not found", status=status.HTTP_404_NOT_FOUND)

        # Create the launcher
        launcher = get_default_client().get_launcher_token(target=serializer.data['target'],
                                                           user=user,
                                                           target_url=serializer.data['target_url'],
                                                           session=session,
                                                           ttl=serializer.data['ttl'])
        resp = self.serializer_class(launcher)
        return Response(resp.data)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango command to generate deployment scripts for a VLE """
from ..base import TeslaDeployCommand
from tesla_ce import models
from tesla_ce import tasks


class Command(TeslaDeployCommand):
    """ Command to rebuild all reports """
    help = 'Rebuild all reports'
    requires_system_checks = '__all__'

    def add_arguments(self, parser):
        """
            Define custom arguments for this command

            :param parser: Input command parser instance
        """
        # Set default arguments
        super().add_arguments(parser)

        parser.add_argument(
            '--course-id',
            help='Course Id',
            default=None,
        )

        parser.add_argument(
            '--institution-id',
            help='Institution Id',
            default=None,
        )

        parser.add_argument(
            '--user-id',
            help='User Id',
            default=None,
        )

    def custom_handle(self):
        """
            Custom actions for this command
        """
        # Get all activities
        activities = models.Activity.objects

        # Filter by institution
        if self._options['institution_id'] is not None:
            activities = activities.filter(vle__institution_id=self._options['institution_id'])

        # Filter by course
        if self._options['course_id'] is not None:
            activities = activities.filter(course_id=self._options['course_id'])

        report_count = 0
        error_count = 0
        for activity in activities.all():
            learners = activity.course.learners
            if self._options['user_id'] is not None:
                learners = learners.filter(id=self._options['user_id'])
            for learner in learners.all():
                try:
                    tasks.reports.update_learner_activity_report(learner.id, activity.id, force_update=True)
                except Exception:
                    error_count += 1
                report_count += 1

        if error_count == 0:
            self.stdout.write(self.style.SUCCESS('{} reports generated'.format(report_count)))
        else:
            self.stdout.write(self.style.ERROR('{} reports generated with {} errors'.format(report_count, error_count)))



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Learner enrolment samples views module """
import json

from django.core.files.storage import default_storage
from rest_framework.views import APIView
from rest_framework.views import Response
from rest_framework.views import status

from ..serializers import StatusSerializer
from ..utils import is_authenticated


class StatusView(APIView):
    """
        Provide an enrolment sample for a learner
    """
    def _get_path_status(self, path):
        if not default_storage.exists(path):
            return "NOT_FOUND", None
        if default_storage.exists('{}.error'.format(path)):
            return "ERROR", json.loads(default_storage.open('{}.error'.format(path),'r').read())
        if default_storage.exists('{}.valid'.format(path)):
            return "VALID", json.loads(default_storage.open('{}.valid'.format(path),'r').read())
        if default_storage.exists('{}.timeout'.format(path)):
            return "TIMEOUT", json.loads(default_storage.open('{}.timeout'.format(path),'r').read())
        return "PENDING", None

    def post(self, request, institution_id,  learner_id, format=None):
        serializer = StatusSerializer(data=request.data, context={'request': request})

        if serializer.is_valid():
            if not is_authenticated(request, institution_id, learner_id, serializer.data):
                return Response(status=status.HTTP_403_FORBIDDEN, data={'errors': "Invalid token"})

            status_response_list = []
            for path in serializer.data['samples']:
                path_status = self._get_path_status(path)
                status_response_list.append({
                    'sample': path,
                    'status': path_status[0],
                    'info': path_status[1]
                })

            return Response(status_response_list)

        return Response(status=status.HTTP_400_BAD_REQUEST, data={'errors': serializer.errors})



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Utility module to perform system checks"""
import hvac
from django.db.utils import ConnectionHandler
from django.db.utils import OperationalError
from requests.exceptions import ConnectionError


def check_vault_connection(config):
    """
        Check HashiCorp Vault status

        :param config: Configuration object
        :return: Report of the status
        :rtype: dict
    """
    # Create a new Vault client
    client = hvac.Client(url=config.get('VAULT_URL'), verify=config.get('VAULT_SSL_VERIFY'))

    # Initialize the report
    report = {}

    # Try to connect vault
    try:
        is_initialized = client.sys.is_initialized()
        report['connected'] = True
        report['initialized'] = is_initialized
    except ConnectionError as e:
        report['connected'] = False
        report['error'] = e

    if report['connected'] and report['initialized']:
        report['sealed'] = client.sys.is_sealed()

    report['ready'] = report['connected'] and report['initialized'] and not report['sealed']

    return report


def check_database_connection(config):
    """
        Check Database status

        :param config: Configuration object
        :return: Report of the status
        :rtype: dict
    """
    # Define the databases
    databases = {
        'admin': {
            'ENGINE': 'django.db.backends.{}'.format(config.get('DB_ENGINE')),
            'HOST': config.get('DB_HOST'),
            'PORT': config.get('DB_PORT'),
            'USER': config.get('DB_ROOT_USER'),
            'PASSWORD': config.get('DB_ROOT_PASSWORD'),
            'NAME': ''
        },
        'default': {
            'ENGINE': 'django.db.backends.{}'.format(config.get('DB_ENGINE')),
            'HOST': config.get('DB_HOST'),
            'PORT': config.get('DB_PORT'),
            'USER': config.get('DB_USER'),
            'PASSWORD': config.get('DB_PASSWORD'),
            'NAME': config.get('DB_NAME'),
        },
    }

    # Create the connections
    connection = ConnectionHandler()
    connection.configure_settings(databases=databases)

    # Check connections
    report = {}
    for db in databases:
        report[db] = {}
        try:
            connection[db].cursor()
            report[db]['connected'] = True
        except OperationalError as e:
            report[db]['connected'] = False
            report[db]['error'] = e
        except Exception as e:
            report[db]['connected'] = False
            report[db]['error'] = e

    return report


def check_vault_configuration(config):
    """
        Check HashiCorp Vault configuration

        :param config: Configuration object
        :return: Report of the configuration check
        :rtype: dict
    """
    valid = True
    errors = []

    # Check URL
    if config.get('VAULT_URL') is None:
        valid = False
        errors.append('Vault URL is not provided')
    elif not config.get('VAULT_URL').startswith('https://') and not config.get('VAULT_URL').startswith('http://'):
        valid = False
        errors.append('Vault URL schema is not provided. Use http or https.')

    # Check connection
    if valid:
        config_report = check_vault_connection(config)
        if not config_report['connected']:
            valid = False
            errors.append('Cannot connect with Vault: {}'.format(config_report['error']))
        else:
            if config_report['initialized']:
                if config.get('VAULT_TOKEN') is None:
                    valid = False
                    errors.append('Vault is initialized, but root token is not provided')
                if config_report['sealed'] and \
                        (config.get('VAULT_KEYS') is None or len(config.get('VAULT_KEYS')) == 0):
                    valid = False
                    errors.append('Vault is sealed, but unseal keys are not provided')
    else:
        config_report = {
            'connected': False
        }

    # Add final status and errors
    config_report['valid'] = valid
    config_report['errors'] = errors

    return config_report


def check_db_configuration(config):
    """
        Check Database configuration

        :param config: Configuration object
        :return: Report of the configuration check
        :rtype: dict
    """
    valid = True
    connected = True
    errors = []
    config_report = {}

    # Check basic configuration
    if config.get('DB_ENGINE') is None:
        valid = False
        errors.append('Database engine not provided')
    if config.get('DB_HOST') is None:
        valid = False
        errors.append('Database host not provided')
    if config.get('DB_PORT') is None:
        valid = False
        errors.append('Database port not provided')
    if (config.get('DB_USER') is None or config.get('DB_PASSWORD') is None) and \
        (config.get('DB_ROOT_USER') is None or config.get('DB_ROOT_PASSWORD') is None):
        valid = False
        errors.append('Database credentials not provided.')
    if valid:
        config_report['databases'] = check_database_connection(config)
        if not config_report['databases']['default']['connected'] and \
                not config_report['databases']['admin']['connected']:
            connected = False
            errors.append('Cannot connect to database.')

    config_report['valid'] = valid
    config_report['connected'] = connected
    config_report['errors'] = errors
    return config_report


def check_tesla_configuration(config):
    """
        Check general TeSLA configuration

        :param config: Configuration object
        :return: Report of the configuration check
        :rtype: dict
    """
    valid = True
    errors = []

    if config.get('TESLA_DOMAIN') is None:
        valid = False
        errors.append('TeSLA domain is missing')

    if config.get('TESLA_ADMIN_MAIL') is None:
        valid = False
        errors.append('TeSLA administrator email is missing')

    return {
        'valid': valid,
        'errors': errors
    }
